from django.contrib.admin import AdminSite
from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.test import TestCase
from django.views import generic

from ambient_toolbox.admin.views.mixins import AdminViewMixin
from ambient_toolbox.tests.mixins import RequestProviderMixin
from testapp.models import MySingleSignalModel


class TestView(AdminViewMixin, generic.TemplateView):
    model = MySingleSignalModel
    admin_page_title = "My fancy title"
    template_name = "testapp/test_template.html"


class AdminViewMixinTest(RequestProviderMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.view = TestView()

        cls.super_user = User.objects.create(username="super_user", is_superuser=True)
        cls.regular_user = User.objects.create(username="test_user", is_superuser=False)

        # View needs a request since django 3.2
        request = cls.get_request(cls.super_user)
        cls.view.request = request

    def test_admin_view_mixin_has_view_permission_positive_case(self):
        self.assertTrue(self.view.has_view_permission(self.super_user))

    def test_admin_view_mixin_has_view_permission_negative_case(self):
        self.assertFalse(self.view.has_view_permission(self.regular_user))

    def test_admin_view_mixin_access_allowed_for_superusers(self):
        request = self.get_request(self.super_user)
        self.view.request = request

        self.view.dispatch(request=request)

    def test_admin_view_mixin_access_blocked_for_non_superusers(self):
        request = self.get_request(self.regular_user)

        with self.assertRaises(PermissionDenied):
            self.view.dispatch(request=request)

    def test_admin_view_mixin_get_admin_site_regular(self):
        self.assertIsInstance(self.view.get_admin_site(), AdminSite)

    def test_admin_view_mixin_get_context_data_regular(self):
        context_data = self.view.get_context_data()

        # Simply assert custom fields are available
        self.assertIn("site_header", context_data)
        self.assertIn("site_title", context_data)
        self.assertIn("name", context_data)
        self.assertIn("original", context_data)
        self.assertIn("is_nav_sidebar_enabled", context_data)
        self.assertIn("available_apps", context_data)
        self.assertIn("opts", context_data)
        self.assertIn("app_label", context_data["opts"])
        self.assertIn("verbose_name", context_data["opts"])
        self.assertIn("verbose_name_plural", context_data["opts"])
        self.assertIn("model_name", context_data["opts"])
        self.assertIn("app_config", context_data["opts"])
        self.assertIn("verbose_name", context_data["opts"]["app_config"])
        self.assertIn("has_permission", context_data)



from django.conf import settings
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from ambient_toolbox.middleware.current_request import CurrentRequestMiddleware


class CreatedAtInfo(models.Model):
    created_at = models.DateTimeField(_("Created at"), default=now, db_index=True)

    class Meta:
        abstract = True

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None, **kwargs):
        # just a fallback for old data
        if not self.created_at:
            self.created_at = now()
        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
            **kwargs,
        )


class CommonInfo(CreatedAtInfo):
    # Automatically add the model's fields to the 'update_fields' list if specified on save()
    ALWAYS_UPDATE_FIELDS = True

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Created by"),
        blank=True,
        null=True,
        related_name="%(app_label)s_%(class)s_created",
        on_delete=models.SET_NULL,
    )
    lastmodified_at = models.DateTimeField(_("Last modified at"), default=now, db_index=True)
    lastmodified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Last modified by"),
        blank=True,
        null=True,
        related_name="%(app_label)s_%(class)s_lastmodified",
        on_delete=models.SET_NULL,
    )

    class Meta:
        abstract = True

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None, **kwargs):
        self.lastmodified_at = now()
        current_user = self.get_current_user()
        self.set_user_fields(current_user)

        # Handle case that somebody only wants to update some fields
        if update_fields is not None and self.ALWAYS_UPDATE_FIELDS:
            update_fields = {"lastmodified_at", "lastmodified_by", "created_at", "created_by"}.union(update_fields)

        super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
            **kwargs,
        )

    @staticmethod
    def get_current_user():
        """
        Get the currently logged-in user over middleware.
        Can be overwritten to use e.g. other middleware or additional functionality.
        :return: user instance
        """
        return CurrentRequestMiddleware.get_current_user()

    def set_user_fields(self, user):
        """
        Set user-related fields before saving the instance.
        If no user with primary key is given the fields are not set.
        :param user: user instance of current user
        """
        if user and user.pk:
            if not self.pk:
                self.created_by = user
            self.lastmodified_by = user



from django.apps import AppConfig
from django.conf import settings
from django.core.checks import Tags, register
from django.utils.translation import gettext_lazy as _


class AmbientToolboxConfig(AppConfig):
    name = "ambient_toolbox"
    verbose_name = _("Ambient Toolbox")


if getattr(settings, "STATIC_ROLE_PERMISSIONS_PATH", None) and not (
    getattr(settings, "STATIC_ROLE_PERMISSIONS_DISABLE_SYSTEM_CHECK", False)
):
    from ambient_toolbox.static_role_permissions.system_check import check_permissions_against_models

    register(check_permissions_against_models, Tags.models)



from typing import Optional

from sentry_sdk.serializer import serialize


class SentryEventScrubber:
    """
    Add this to the before_send and the before_send_transaction hook in order to remove all sensitive data from
    a Sentry event. It will also remove sensitive data in stacktraces.

    Usage:
    before_send = SentryEventScrubber().scrub_sensitive_data_from_sentry_event
    before_send_transaction = SentryEventScrubber().scrub_sensitive_data_from_sentry_event
    """

    def __init__(self, denylist: Optional[list[str]] = None, standard_denylist: Optional[bool] = True) -> None:
        """
        Arguments:
        ---------
        denylist: A list of keys that should be scrubbed from the Sentry event.
        standard_denylist: By default, certain keys are already scrubbed from the event.

        """
        self.denylist = [] if denylist is None else denylist
        self.standard_denylist = (
            [
                "username",
                "email",
                "ip_address",
                "serializer",
                "admin",
            ]
            if standard_denylist
            else []
        )

    def scrub_sensitive_data_from_sentry_event(self, event, _hint):
        from sentry_sdk.scrubber import DEFAULT_DENYLIST, EventScrubber

        EventScrubber(denylist=list(set(DEFAULT_DENYLIST + self.standard_denylist + self.denylist))).scrub_event(event)
        return serialize(event)


def strip_sensitive_data_from_sentry_event(event, hint):
    """
    A more lightweight way of cleaning sensitive data from the sentry event.

    Helper method to strip sensitive user data from default sentry event when "send_default_pii" is set to True.
    All user-related data except the internal user id will be removed.
    Variable "hint" contains information about the error itself which we don't need here.
    Requires "sentry-sdk>=1.5.0" to work.
    """
    try:
        del event["user"]["username"]
    except KeyError:
        pass
    try:
        del event["user"]["email"]
    except KeyError:
        pass
    try:
        del event["user"]["ip_address"]
    except KeyError:
        pass
    return event



from django.db.models import QuerySet

from ambient_toolbox.selectors.base import Selector


class AbstractUserSpecificSelectorMixin:
    """
    Abstract selector mixin to inherit from to implement a basic permission pattern.
    Refer to the documentation for further details.
    """

    def visible_for(self, user_id: int) -> QuerySet:
        raise NotImplementedError

    def editable_for(self, user_id: int) -> QuerySet:
        raise NotImplementedError

    def deletable_for(self, user_id: int) -> QuerySet:
        raise NotImplementedError


class GloballyVisibleSelector(AbstractUserSpecificSelectorMixin, Selector):
    """
    Selector for classes which do NOT have any visibility restrictions. Use with caution!
    """

    def visible_for(self, user_id: int) -> QuerySet:
        return self.all()

    def editable_for(self, user_id: int) -> QuerySet:
        return self.visible_for(user_id=user_id)

    def deletable_for(self, user_id: int) -> QuerySet:
        return self.visible_for(user_id=user_id)



from django.conf import settings

# Test validator
TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST = []
try:
    TEST_STRUCTURE_VALIDATOR_BASE_DIR = settings.BASE_DIR
except AttributeError:
    TEST_STRUCTURE_VALIDATOR_BASE_DIR = ""
TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME = "apps"
TEST_STRUCTURE_VALIDATOR_APP_LIST = settings.INSTALLED_APPS
TEST_STRUCTURE_VALIDATOR_IGNORED_DIRECTORY_LIST = []



import re

from django.conf import settings
from django.core.mail.backends.smtp import EmailBackend as SMTPEmailBackend


class WhitelistEmailBackend(SMTPEmailBackend):
    """
    Via the following settings it is possible to configure if mails are sent to all domains.
    If not, you can configure a redirect to an inbox via CATCHALL.

    EMAIL_BACKEND = 'ambient_toolbox.mail.backends.whitelist_smtp.WhitelistEmailBackend'
    EMAIL_BACKEND_DOMAIN_WHITELIST = ['ambient.digital']
    EMAIL_BACKEND_REDIRECT_ADDRESS = '%s@testuser.ambient.digital'

    If `EMAIL_BACKEND_REDIRECT_ADDRESS` is set, a mail to `john.doe@example.com` will be redirected to
    `john.doe_example.com@testuser.ambient.digital`
    """

    @staticmethod
    def get_domain_whitelist() -> list:
        """
        Getter for configuration variable from the settings.
        Will return a list of domains: ['ambient.digital', 'ambient.digital']
        """
        return getattr(settings, "EMAIL_BACKEND_DOMAIN_WHITELIST", [])

    @staticmethod
    def get_email_regex():
        """
        Getter for configuration variable from the settings.
        Will return a RegEX to match email whitelisted domains.
        """
        return r"^[\w\-\.]+@(%s)$" % "|".join(x for x in WhitelistEmailBackend.get_domain_whitelist()).replace(
            ".", r"\."
        )

    @staticmethod
    def get_backend_redirect_address() -> str:
        """
        Getter for configuration variable from the settings.
        Will return a string with a placeholder for redirecting non-whitelisted domains.
        """
        return settings.EMAIL_BACKEND_REDIRECT_ADDRESS

    @staticmethod
    def whitify_mail_addresses(mail_address_list: list) -> list:
        """
        Check for every recipient in the list if its domain is included in the whitelist.
        If not, and we have a redirect address configured, we change the original mail address to something new,
        according to our configuration.
        """
        allowed_recipients = []
        for to in mail_address_list:
            if re.search(WhitelistEmailBackend.get_email_regex(), to):
                allowed_recipients.append(to)
            elif WhitelistEmailBackend.get_backend_redirect_address():
                # Send not allowed emails to the configured redirect address (with CATCHALL)
                allowed_recipients.append(WhitelistEmailBackend.get_backend_redirect_address() % to.replace("@", "_"))
        return allowed_recipients

    def _process_recipients(self, email_messages):
        """
        Helper method to wrap custom logic of this backend. Required to make it testable.
        """
        for email in email_messages:
            allowed_recipients = self.whitify_mail_addresses(email.to)
            email.to = allowed_recipients
        return email_messages

    def send_messages(self, email_messages):
        """
        Checks if email-recipients are in allowed domains and cancels if not.
        Uses regular smtp-sending afterward.
        """
        email_messages = self._process_recipients(email_messages)
        return super().send_messages(email_messages)



from django.db.models import manager


class Selector(manager.Manager):
    """
    A base class for query selectors. Please refer to the docs for further enlightenment how this novel
    pattern works.
    It's derived from the manager to use Django's magic to inject the current class into the "model" attribute.
    """

    model = None



from argparse import ArgumentParser
from unittest import mock

from django.test import TestCase
from django.test.utils import override_settings

from ambient_toolbox.management.commands.install_permission_fixtures import Command
from ambient_toolbox.permissions.fixtures.services import PermissionSetupService


class InstallPermissionFixturesCommandTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

    def test_add_arguments_regular(self):
        command = Command()

        parser = ArgumentParser()

        command.add_arguments(parser)

        self.assertTrue("--dry-run" in [action.option_strings[0] for action in parser._actions])

    @override_settings(GROUP_PERMISSION_FIXTURES=["testapp.permissions.TestGroupDeclaration"])
    @mock.patch.object(PermissionSetupService, "process", return_value=([], []))
    def test_run_command_regular(self, mocked_process):
        command = Command()
        command.handle()

        mocked_process.assert_called_once()

    @mock.patch.object(PermissionSetupService, "process")
    def test_run_command_no_settings_variable(self, mocked_process):
        command = Command()
        command.handle()

        mocked_process.assert_not_called()

    @mock.patch.object(PermissionSetupService, "process")
    def test_run_command_dry_run(self, mocked_process):
        command = Command()
        command.handle(dry_run=True)

        mocked_process.assert_not_called()



from .cache import *  # noqa: F403
from .date import *  # noqa: F403
from .log_whodid import *  # noqa: F403
from .math import *  # noqa: F403
from .model import *  # noqa: F403
from .named_tuple import *  # noqa: F403



from django.test import TestCase

from ambient_toolbox.templatetags.ai_number_tags import currency, divide, multiply, subtract, to_int


class AiNumberTagTest(TestCase):
    def test_multiply_regular(self):
        self.assertEqual(multiply(3, 3), 9)

    def test_multiply_string(self):
        self.assertEqual(multiply("3", 3), 9)

    def test_multiply_german_float(self):
        self.assertEqual(multiply("3,19", 3), 9.57)

    def test_multiply_no_value(self):
        self.assertIsNone(multiply(None, 3))

    def test_subtract_regular(self):
        self.assertEqual(subtract(4, 1), 3)

    def test_divide_regular(self):
        self.assertEqual(divide(8, 2), 4)

    def test_divide_no_value(self):
        self.assertEqual(divide(None, 2), None)

    def test_to_int_case_number(self):
        self.assertEqual(to_int("42"), 42)

    def test_to_int_case_text(self):
        self.assertEqual(to_int("Aubrey"), 0)

    def test_currency_regular(self):
        self.assertEqual(currency(12.4), "12,40€")

    def test_currency_no_value(self):
        self.assertEqual(currency(None), "-")



from ambient_package_update.metadata.author import PackageAuthor
from ambient_package_update.metadata.constants import (
    DEV_DEPENDENCIES,
    LICENSE_MIT,
    SUPPORTED_DJANGO_VERSIONS,
    SUPPORTED_PYTHON_VERSIONS,
)
from ambient_package_update.metadata.maintainer import PackageMaintainer
from ambient_package_update.metadata.package import PackageMetadata
from ambient_package_update.metadata.readme import ReadmeContent
from ambient_package_update.metadata.ruff_ignored_inspection import RuffIgnoredInspection

METADATA = PackageMetadata(
    package_name="ambient_toolbox",
    github_package_group="ambient-innovation",
    company="Ambient Innovation: GmbH",
    authors=[
        PackageAuthor(
            name="Ambient Digital",
            email="hello@ambient.digital",
        ),
    ],
    maintainer=PackageMaintainer(name="Ambient Digital", url="https://ambient.digital/", email="hello@ambient.digital"),
    min_coverage=86.29,
    development_status="5 - Production/Stable",
    license=LICENSE_MIT,
    license_year=2012,
    readme_content=ReadmeContent(
        tagline="Python toolbox of Ambient Digital containing an abundance of useful tools and gadgets.",
    ),
    dependencies=[
        f"Django>={SUPPORTED_DJANGO_VERSIONS[0]}",
        "python-dateutil>=2.5.3",
        # We keep this until we drop Python 3.8
        "pytz",
    ],
    supported_django_versions=SUPPORTED_DJANGO_VERSIONS,
    supported_python_versions=SUPPORTED_PYTHON_VERSIONS,
    has_migrations=True,
    optional_dependencies={
        "dev": [
            *DEV_DEPENDENCIES,
            "gevent~=23.9",
            "httpx~=0.27",
        ],
        "drf": [
            "djangorestframework>=3.8.2",
        ],
        "graphql": [
            "graphene-django>=2.2.0",
            "django-graphql-jwt>=0.2.1",
        ],
        "bleacher": [
            "nh3>=0.2,<1",
        ],
        "sentry": [
            "sentry-sdk>=1.19.1",
        ],
        "view-layer": [
            "django-crispy-forms>=1.4.0",
        ],
    },
    ruff_ignore_list=[
        RuffIgnoredInspection(key="N999", comment="Project name contains underscore, not fixable"),
        RuffIgnoredInspection(key="A003", comment="Django attributes shadow python builtins"),
        RuffIgnoredInspection(key="DJ001", comment="Django model text-based fields shouldn't be nullable"),
        RuffIgnoredInspection(key="B905", comment="Can be enabled when Python <=3.9 support is dropped"),
        RuffIgnoredInspection(key="DTZ001", comment='TODO will affect "tz_today()" method'),
        RuffIgnoredInspection(key="DTZ005", comment='TODO will affect "tz_today()" method'),
        RuffIgnoredInspection(
            key="RUF012", comment="Mutable class attributes should be annotated with `typing.ClassVar`"
        ),
        RuffIgnoredInspection(key="TD002", comment="Missing author in TODO"),
        RuffIgnoredInspection(key="TD003", comment="Missing issue link on the line following this TODO"),
        RuffIgnoredInspection(key="TRY002", comment="Checks for code that raises Exception directly."),
        RuffIgnoredInspection(
            key="TRY003",
            comment="Checks for long exception messages that are not defined in the exception class itself.",
        ),
    ],
)



import threading
from http import HTTPStatus
from unittest.mock import Mock

from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse
from django.test import TestCase

from ambient_toolbox.middleware.current_user import CurrentUserMiddleware


class CurrentUserMiddlewareTest(TestCase):
    def test_current_user_is_none_if_no_user_given(self):
        self.assertIsNone(CurrentUserMiddleware.get_current_user())

    def test_current_user_is_none_if_request_user_is_none(self):
        response = set_current_user(user=None)
        self.assertEqual(response.status_code, HTTPStatus.NO_CONTENT)

    def test_current_user_is_same_as_request_user(self):
        new_user = Mock(user_name="test_user")
        response = set_current_user(user=new_user)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_current_user_is_thread_safe(self):
        user1 = Mock(user_name="user1")
        user2 = Mock(user_name="user2")
        current_users = []
        ready_event = threading.Event()
        proceed_event = threading.Event()
        first_thread = threading.Thread(
            target=set_current_user,
            args=(user1, current_users, ready_event, proceed_event),
        )
        second_thread = threading.Thread(
            target=set_current_user,
            args=(user2, current_users),
        )
        first_thread.start()
        ready_event.wait()
        second_thread.start()
        second_thread.join()
        proceed_event.set()
        first_thread.join()
        self.assertEqual(current_users[0], user2)
        self.assertEqual(current_users[0].user_name, "user2")
        self.assertEqual(current_users[1], user1)
        self.assertEqual(current_users[1].user_name, "user1")

    def test_user_is_cleared_after_request(self):
        user = Mock(user_name="test_user")
        request = Mock(user=user)
        middleware = CurrentUserMiddleware(get_response=lambda request: HttpResponse(status=HTTPStatus.OK))
        response = middleware(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertIsNone(CurrentUserMiddleware.get_current_user())

    def test_replaced_user_is_reflected_in_middleware(self):
        # This is roughly what happens when using eg. DRF token auth: during
        # middleware processing, request.user is an AnonymousUser, and at some
        # later point DRF replaces it with a user determined from given auth
        # info (eg. a token in the request header).
        # This should ideally be taken into account in our middleware since it
        # is also used to provide the user for `CommonInfo.lastmodified_by`.
        def get_response(request):
            replaced_user = Mock(user_name="replaced_user")
            request.user = replaced_user
            user_from_mw = CurrentUserMiddleware.get_current_user()
            if user_from_mw is not replaced_user:
                return HttpResponse(status=HTTPStatus.CONFLICT)  # pragma: no cover
            return HttpResponse(status=HTTPStatus.OK)

        middleware = CurrentUserMiddleware(get_response)
        anonymous_user = AnonymousUser()
        request = Mock(user=anonymous_user)
        response = middleware(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)


def set_current_user(user=None, current_users=None, ready_event=None, proceed_event=None):
    def get_response(request):
        if ready_event is not None:
            ready_event.set()
        if proceed_event is not None:
            proceed_event.wait()
        user_from_mw = CurrentUserMiddleware.get_current_user()
        if current_users is not None:
            current_users.append(user_from_mw)
        if user_from_mw is not request.user:
            return HttpResponse(status=HTTPStatus.CONFLICT)  # pragma: no cover
        if user_from_mw is None:
            return HttpResponse(status=HTTPStatus.NO_CONTENT)
        return HttpResponse(status=HTTPStatus.OK)

    middleware = CurrentUserMiddleware(get_response)
    request = Mock()
    request.user = user
    return middleware(request)



from unittest import mock

from django.contrib import admin
from django.contrib.admin import AdminSite, ModelAdmin
from django.contrib.auth.models import User
from django.core.handlers.wsgi import WSGIRequest
from django.test import RequestFactory, TestCase

from ambient_toolbox.admin.model_admins.classes import EditableOnlyAdmin, ReadOnlyAdmin
from ambient_toolbox.tests.mixins import RequestProviderMixin
from testapp.models import MyMultipleSignalModel, MySingleSignalModel


class TestReadOnlyAdmin(ReadOnlyAdmin):
    model = MySingleSignalModel


class TestEditableOnlyAdmin(EditableOnlyAdmin):
    model = MySingleSignalModel


class AdminClassesTest(RequestProviderMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.super_user = User.objects.create(username="super_user", is_superuser=True)

        admin.site.register(MySingleSignalModel, TestReadOnlyAdmin)
        admin.site.register(MyMultipleSignalModel, TestEditableOnlyAdmin)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        admin.site.unregister(MySingleSignalModel)
        admin.site.unregister(MyMultipleSignalModel)

    def test_read_only_admin_all_fields_readonly(self):
        obj = MySingleSignalModel(value=1)

        admin_class = TestReadOnlyAdmin(model=obj, admin_site=admin.site)
        readonly_fields = admin_class.get_readonly_fields(request=self.get_request(), obj=obj)

        self.assertEqual(len(readonly_fields), 2)
        self.assertIn("id", readonly_fields)
        self.assertIn("value", readonly_fields)

    def test_read_only_admin_no_change_permissions(self):
        admin_class = TestReadOnlyAdmin(model=MySingleSignalModel, admin_site=admin.site)

        request = self.get_request(self.super_user)

        self.assertFalse(admin_class.has_add_permission(request))
        self.assertFalse(admin_class.has_change_permission(request))
        self.assertFalse(admin_class.has_delete_permission(request))

    def test_editable_only_admin_delete_action_removed(self):
        obj = MyMultipleSignalModel(value=1)
        admin_class = TestEditableOnlyAdmin(model=obj, admin_site=admin.site)

        request = self.get_request(self.super_user)
        actions = admin_class.get_actions(request=request)

        self.assertNotIn("delete_selected", actions)

    def test_editable_only_admin_no_change_permissions(self):
        admin_class = TestEditableOnlyAdmin(model=MyMultipleSignalModel, admin_site=admin.site)

        request = self.get_request(self.super_user)

        self.assertTrue(admin_class.has_change_permission(request))

        self.assertFalse(admin_class.has_add_permission(request))
        self.assertFalse(admin_class.has_delete_permission(request))


class ReadOnlyAdminTest(TestCase):
    user: User
    request: WSGIRequest

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(username="testuser", password="testpassword", is_superuser=True)

        factory = RequestFactory()
        cls.request = factory.get(f"/admin/auth/user/{cls.user.id}/change/")
        cls.request.user = cls.user

    def test_changeform_view_regular(self):
        model_admin = ReadOnlyAdmin(User, AdminSite())
        response = model_admin.changeform_view(self.request, str(self.user.id))
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "Save and continue editing")
        self.assertNotContains(response, "Save")

    def test_has_add_permission_regular(self):
        model_admin = ReadOnlyAdmin(User, AdminSite())
        self.assertFalse(model_admin.has_add_permission(self.request))

    def test_has_change_permission_regular(self):
        model_admin = ReadOnlyAdmin(User, AdminSite())
        self.assertFalse(model_admin.has_change_permission(self.request))

    def test_has_delete_permission_regular(self):
        model_admin = ReadOnlyAdmin(User, AdminSite())
        self.assertFalse(model_admin.has_delete_permission(self.request))


class EditableOnlyAdminTest(TestCase):
    user: User
    request: WSGIRequest

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(username="testuser", password="testpassword", is_superuser=True)

        factory = RequestFactory()
        cls.request = factory.get(f"/admin/auth/user/{cls.user.id}/change/")
        cls.request.user = cls.user

    @mock.patch.object(ModelAdmin, "get_actions", return_value={"delete_selected": 1})
    def test_get_actions_regular(self, *args):
        model_admin = EditableOnlyAdmin(User, AdminSite())
        actions = model_admin.get_actions(self.request)
        self.assertIsInstance(actions, dict)
        self.assertNotIn("delete_selected", actions)



from django.contrib.auth.models import Permission
from django.test import TestCase

from ambient_toolbox.utils import get_cached_related_obj, object_to_dict
from testapp.models import ForeignKeyRelatedModel, MySingleSignalModel


class UtilModelTest(TestCase):
    def test_object_to_dict_regular(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj), {"value": obj.value})

    def test_object_to_dict_blacklist(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj, ["value"]), {})

    def test_object_to_dict_with_id_with_blacklist(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj, ["value"], True), {"id": obj.id})

    def test_with_id_no_blacklist(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj, include_id=True), {"id": obj.id, "value": obj.value})

    def test_object_to_dict_valid_fields_append(self):
        obj = MySingleSignalModel.objects.create(value=17)
        dummy_instance = ForeignKeyRelatedModel(single_signal=obj)

        valid_data = object_to_dict(dummy_instance)

        self.assertIn("single_signal_id", valid_data)
        self.assertEqual(valid_data["single_signal_id"], obj.id)

    # --- Test get_cached_related_obj (happy path) ---
    def test_get_cached_related_obj_with_select_related(self):
        permission_obj = Permission.objects.select_related("content_type").first()
        with self.assertNumQueries(0):
            content_type = get_cached_related_obj(permission_obj, "content_type")
            self.assertIsNotNone(content_type)

    def test_get_cached_related_obj_with_prefetch_related(self):
        permission_obj = Permission.objects.prefetch_related("content_type").first()
        with self.assertNumQueries(0):
            content_type = get_cached_related_obj(permission_obj, "content_type")
            self.assertIsNotNone(content_type)

    def test_get_cached_related_obj_without_cached_and_silently_return_none(self):
        permission_obj = Permission.objects.first()
        with self.assertNumQueries(0):
            content_type = get_cached_related_obj(permission_obj, "content_type", silently_return_none=True)
            self.assertIsNone(content_type)

    # --- Test get_cached_related_obj (error path) ---

    def test_get_cached_related_obj_without_cached(self):
        permission_obj = Permission.objects.first()
        with self.assertNumQueries(0):
            with self.assertRaises(AttributeError) as ctx:
                get_cached_related_obj(permission_obj, "content_type")
            self.assertEqual(
                f'Field "content_type" not found in `fields_cache` of {permission_obj.__class__.__name__} object. '
                "Did you forget to use `select_related()` or `prefetch_related()`?",
                str(ctx.exception),
            )

    def test_get_cached_related_obj_with_non_existing_field(self):
        permission_obj = Permission.objects.first()
        with self.assertNumQueries(0):
            with self.assertRaises(AttributeError) as ctx:
                get_cached_related_obj(permission_obj, "foo")
            self.assertEqual(
                f'Field "foo" not found in `fields_cache` of {permission_obj.__class__.__name__} object. '
                "Did you forget to use `select_related()` or `prefetch_related()`?",
                str(ctx.exception),
            )

    def test_get_cached_related_obj_with_no_fk_field(self):
        permission_obj = Permission.objects.first()
        with self.assertNumQueries(0):
            with self.assertRaises(AttributeError) as ctx:
                get_cached_related_obj(permission_obj, "name")
            self.assertEqual(
                f'Field "name" not found in `fields_cache` of {permission_obj.__class__.__name__} object. '
                "Did you forget to use `select_related()` or `prefetch_related()`?",
                str(ctx.exception),
            )



from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings

from ambient_toolbox.validators.auth_password.special_chars import SpecialCharValidator


class SpecialCharValidatorTest(TestCase):
    def test_validate_happy_path(self):
        validator = SpecialCharValidator()
        self.assertIsNone(validator.validate("Admin0404!"))

    def test_validate_missing_special_char(self):
        validator = SpecialCharValidator()
        with self.assertRaisesMessage(
            ValidationError, 'The password has to contain one of the following special characters: "@#$%!^&*"'
        ):
            validator.validate("EasyPassword")

    def test_get_help_text_regular(self):
        validator = SpecialCharValidator()
        self.assertEqual(
            validator.get_help_text(), 'The password has to contain one of the following special characters: "@#$%!^&*"'
        )

    @override_settings(AUTH_PASSWORD_VALIDATORS=["ambient_toolbox.validators.SpecialCharValidator"])
    def test_functional_happy_path(self):
        user = User()
        self.assertIsNone(user.set_password("Admin0404!"))

    @override_settings(AUTH_PASSWORD_VALIDATORS=["ambient_toolbox.validators.auth_password.SpecialCharValidator"])
    def test_functional_special_char_missing(self):
        user = User()
        self.assertFalse(user.set_password("EasyPassword"))



from django.shortcuts import render


class _FormsetMixin:
    """
    Mixin for handling a django form with a formset. Extends the form handling logic to handle a formset object.

    Attention: Do NOT use directly. Do use FormsetUpdateViewMixin or FormsetCreateViewMixin
    """

    formset_class = None
    object = None

    def get_formset_kwargs(self):
        # may be overridden or extended
        return {"instance": self.object}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["formset"] = self.formset_class(**self.get_formset_kwargs())
        return context

    def form_valid(self, form, formset):
        # Updates `self.object` internally and returns a redirect response instance
        response = super().form_valid(form=form)

        # Update formset
        formset.instance = self.object
        formset.save()

        if hasattr(self, "additional_is_valid"):
            self.additional_is_valid(form, formset)

        # Return response (a redirect)
        return response

    def post(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        form = form_class(**self.get_form_kwargs())
        formset = self.formset_class(request.POST, request.FILES, **self.get_formset_kwargs())

        # Form and formset valid?
        if form.is_valid() and formset.is_valid():
            return self.form_valid(form=form, formset=formset)

        # Get all context data
        context = self.get_context_data()

        # Update form and formset variables
        context["form"] = form
        context["formset"] = formset

        # Pass all data to template
        return render(request, self.get_template_names(), context)


class FormsetUpdateViewMixin(_FormsetMixin):
    """
    Can be used to validate and save a formset.
    Usage is similar to regular UpdateView:

    class MyModelUpdateView(FormsetUpdateViewMixin, generic.UpdateView):
        form_class = MyModelSettingsForm
        template_name = 'myapp/my_model_edit.html'
        formset_class = inlineformset_factory(MyModel, MyFkRelatedModel, form=MyFkRelatedModelForm, extra=0, \
                                              can_delete=False)
        model = MyModel
        success_url = reverse_lazy('my_model:edit')

        def additional_is_valid(self, form, formset):
            messages.add_message(self.request, messages.INFO, 'Update was successful.')
    """

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)


class FormsetCreateViewMixin(_FormsetMixin):
    """
    Can be used to validate and save a formset.
    Usage is similar to regular CreateView. See FormsetUpdateViewMixin.
    """

    def post(self, request, *args, **kwargs):
        self.object = None
        return super().post(request, *args, **kwargs)



from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration, PermissionModelDeclaration


class TestGroupDeclaration(GroupPermissionDeclaration):
    name = ("group_1",)
    permission_list = (
        PermissionModelDeclaration(
            app_label="testapp",
            codename_list=["view_mysinglesignalmodel"],
            model="mysinglesignalmodel",
        ),
    )



from rest_framework import serializers

from testapp.models import MySingleSignalModel


class MySingleSignalModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = MySingleSignalModel
        fields = (
            "id",
            "value",
        )



from django.contrib import messages
from django.test import TestCase

from ambient_toolbox.tests.mixins import DjangoMessagingFrameworkTestMixin, RequestProviderMixin


class DjangoMessagingFrameworkTestMixinTest(RequestProviderMixin, DjangoMessagingFrameworkTestMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.request = cls.get_request()

    def test_full_message_found(self):
        messages.add_message(self.request, messages.INFO, "My message")
        self.assert_full_message_in_request(request=self.request, message="My message")

    def test_partial_message_found(self):
        messages.add_message(self.request, messages.INFO, "My message")
        self.assert_partial_message_in_request(request=self.request, message="My")

    def test_message_not_found(self):
        messages.add_message(self.request, messages.INFO, "My message")
        self.assert_message_not_in_request(request=self.request, message="Ninja")

    def test_message_not_found_but_found(self):
        messages.add_message(self.request, messages.INFO, "My message")
        with self.assertRaises(AssertionError):
            self.assert_message_not_in_request(request=self.request, message="My message")



from django import template
from django.template.defaultfilters import stringfilter

register = template.Library()


@register.filter(name="get_first_char")
def get_first_char(value):
    """
    Returns the first char of the given string
    :param value:
    :return:
    """
    return value[:1]


@register.filter(name="concat")
def concat(obj, value: str) -> str:
    """
    Concatenates the two given strings
    """
    return f"{obj}{value}"


@register.filter
@stringfilter
def trim(value):
    """
    Strips the whitespaces of the given value

    :param value:
    :return:
    """
    return value.strip()



from django.contrib.auth.models import User
from django.test import TestCase
from django.views.generic import View
from django.views.generic.detail import SingleObjectMixin

from ambient_toolbox.tests.mixins import RequestProviderMixin
from ambient_toolbox.view_layer.views import ToggleView
from testapp.views import UserInFormKwargsMixinView


class UserInFormKwargsMixinTest(RequestProviderMixin, TestCase):
    def test_get_form_kwargs_regular(self):
        user = User(username="my-user")

        view = UserInFormKwargsMixinView()
        view.request = self.get_request(user=user)
        form_kwargs = view.get_form_kwargs()

        self.assertIn("user", form_kwargs)
        self.assertEqual(form_kwargs["user"], user)


class ToggleViewTest(RequestProviderMixin, TestCase):
    def test_http_method_set_correctly(self):
        self.assertEqual(ToggleView.http_method_names, ("post",))

    def test_post_raises_not_implemented_error(self):
        with self.assertRaises(NotImplementedError):
            view = ToggleView()
            view.post(request=self.get_request())

    def test_class_inherits_from_single_object_mixin(self):
        self.assertTrue(issubclass(ToggleView, SingleObjectMixin))

    def test_class_inherits_from_generic_view(self):
        self.assertTrue(issubclass(ToggleView, View))



from pydoc import locate

from django.conf import settings
from django.core.management.base import BaseCommand

from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration
from ambient_toolbox.permissions.fixtures.services import PermissionSetupService


class Command(BaseCommand):
    help = "Installs Ambient toolbox improved group permission fixtures"

    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Doesn't persist any changes in the database",
        )

    def handle(self, *args, **options):
        dry_run = options.get("dry_run")
        if dry_run:
            print('Starting in "dry-run" mode...')

        try:
            fixture_declaration_list = settings.GROUP_PERMISSION_FIXTURES
        except AttributeError:
            print("No fixtures found in Django settings.")
            fixture_declaration_list = []

        for declaration_path in fixture_declaration_list:
            print(f'Reading fixture declaration "{declaration_path}"...')
            declaration_class: GroupPermissionDeclaration = locate(declaration_path)

            assert isinstance(
                declaration_class, type(GroupPermissionDeclaration)
            ), f'Could\'t load group declaration "{declaration_path}".'

            print(f'> Installing permissions of group "{declaration_class.name}"...')
            service = PermissionSetupService(group_declaration=declaration_class, dry_run=dry_run)
            new_permissions, removed_permissions = service.process()

            print(f"> Newly installed permissions: {new_permissions}")
            print(f"> Removed permissions: {removed_permissions}\n")



from unittest import mock

from django.test import SimpleTestCase

from ambient_toolbox.gitlab.coverage import CoverageService


@mock.patch.dict("os.environ", {"CI_PIPELINE_ID": "17", "CI_PROJECT_ID": "27"})
class CoverageServiceTest(SimpleTestCase):
    def test_get_disable_coverage_integer_false(self):
        service = CoverageService()
        result = service.get_disable_coverage(disable_env="0")

        self.assertIsInstance(result, bool)
        self.assertFalse(result)

    def test_get_disable_coverage_integer_true(self):
        service = CoverageService()
        result = service.get_disable_coverage(disable_env="1")

        self.assertIsInstance(result, bool)
        self.assertTrue(result)

    def test_get_disable_coverage_string_var_bool_false(self):
        service = CoverageService()
        result = service.get_disable_coverage(disable_env="False")

        self.assertIsInstance(result, bool)
        self.assertFalse(result)

    def test_get_disable_coverage_string_var_bool_true(self):
        service = CoverageService()
        result = service.get_disable_coverage(disable_env="True")

        self.assertIsInstance(result, bool)
        self.assertTrue(result)

    def test_get_disable_coverage_string_var_random(self):
        service = CoverageService()
        result = service.get_disable_coverage(disable_env="Wololo")

        self.assertIsInstance(result, bool)
        self.assertTrue(result)



from django.db.models import manager
from django.test import TestCase

from ambient_toolbox.selectors.base import Selector
from testapp.models import ModelWithSelector


class SelectorTest(TestCase):
    def test_selector_inherits_from_django_manager(self):
        base_selector = Selector()
        self.assertIsInstance(base_selector, manager.Manager)

    def test_registering_in_model_is_possible(self):
        self.assertIsInstance(ModelWithSelector.selectors, manager.Manager)



import warnings

import nh3


class BleacherMixin:
    """
    Removes HTML tags and attributes from the fields defined in :py:attr:`BLEACH_FIELD_LIST`.
    Allowed tags and attributes are defined in the :py:attr:`ALLOWED_TAGS` and :py:attr:`ALLOWED_ATTRIBUTES` attributes.
    If :py:attr:`ALLOWED_TAGS` and :py:attr:`ALLOWED_ATTRIBUTES` are not specified on a model a set of default value is
    used:

    ALLOWED_TAGS:
      * span
      * p
      * h1
      * h2
      * h3
      * h4
      * h5
      * h6
      * img
      * div
      * u
      * br
      * blockquote

    ALLOWED_ATTRIBUTES:
      * all tags: class, style, id
      * a: href, rel
      * img: alt, src
    """

    BLEACH_FIELD_LIST: list[str] = []

    DEFAULT_ALLOWED_ATTRIBUTES: dict[str, set[str]] = {
        **nh3.ALLOWED_ATTRIBUTES,
        "*": ("class", "style", "id"),
    }

    DEFAULT_ALLOWED_TAGS: list[str] = [
        *nh3.ALLOWED_TAGS,
        "span",
        "p",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "img",
        "div",
        "u",
        "br",
        "blockquote",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields_to_bleach = getattr(self, "BLEACH_FIELD_LIST", [])
        self.allowed_tags = set(getattr(self, "ALLOWED_TAGS", self.DEFAULT_ALLOWED_TAGS))
        self.allowed_attributes = getattr(self, "ALLOWED_ATTRIBUTES", self.DEFAULT_ALLOWED_ATTRIBUTES)

        for tag, attribute_list in self.allowed_attributes.items():
            if isinstance(attribute_list, (list, tuple)):
                self.allowed_attributes[tag] = set(attribute_list)
                warnings.warn(
                    "Please use a set instead of a list for the BleacherMixin.ALLOWED_ATTRIBUTES attribute.",
                    category=DeprecationWarning,
                    stacklevel=1,
                )

    def _bleach_field(self, field_name):
        str_to_bleach = getattr(self, field_name, "")
        if str_to_bleach:
            cleaned_value = nh3.clean(
                str_to_bleach,
                tags=self.allowed_tags,
                attributes=self.allowed_attributes,
            )
            setattr(self, field_name, cleaned_value)

    def save(self, *args, **kwargs):
        for field in self.fields_to_bleach:
            self._bleach_field(field)

        super().save(*args, **kwargs)



from django.contrib import admin
from django.urls import include, path
from django.views.generic import TemplateView
from rest_framework import routers

from testapp.api.urls import model_router

router = routers.DefaultRouter()
router.registry.extend(model_router.registry)

urlpatterns = [
    # django Admin
    path("admin/", admin.site.urls),
    # REST Viewsets
    path("api/v1/", include(router.urls)),
    # Custom login view
    path("other/login/", TemplateView.as_view(template_name="testapp/test_template.html"), name="other-login-view"),
]



from django.db.models.signals import post_save, pre_save


class PermissionModelMixin:
    """
    Abstract model class to put permissions in which don't belong to a "real" database model.
    Take care that you have to set "Meta.default_permissions" to an empty tuple/list if you want to avoid having the
    Django default permissions created.
    Inspiration / Source: https://stackoverflow.com/a/37988537/1331671
    """

    class Meta:
        # No database table creation or deletion operations will be performed for this model.
        managed = False


class SaveWithoutSignalsMixin:
    """
    Mixin to provide a save method that temporarily unhooks all signals and restores them after save/error
    """

    def save_without_signals(self, *args, **kwargs):
        # Temporarily store signal receivers to restore them later
        pre_save_receivers = pre_save.receivers
        post_save_receivers = post_save.receivers

        # Clear signal receivers to make sure, that no signal is triggered when .save() is called
        pre_save.receivers = []
        post_save.receivers = []

        # Save without any signals
        instance = self.save(*args, **kwargs)

        # Restore signal receivers
        pre_save.receivers = pre_save_receivers
        post_save.receivers = post_save_receivers

        return instance



from unittest import mock

from django.contrib.auth.models import AnonymousUser, User
from django.test import TestCase, override_settings

from ambient_toolbox.static_role_permissions.permissions import load_static_role_permissions


@override_settings(
    AUTHENTICATION_BACKENDS=[
        "ambient_toolbox.static_role_permissions.auth_backend.StaticRolePermissionBackend",
    ]
)
class StaticRolePermissionTests(TestCase):
    @override_settings(STATIC_ROLE_PERMISSIONS_PATH="tests.static_role_permissions.dummy_permissions.PERMISSIONS_DICT")
    def test_load_permissions_from_settings_path(self):
        permissions = load_static_role_permissions()
        self.assertEqual(
            {"role_1": {"auth.add_user", "auth.view_user"}, "role_2": {}},
            permissions,
        )

    def test_load_permissions_from_settings_path_without_path_set(self):
        with self.assertRaisesMessage(AssertionError, "STATIC_ROLE_PERMISSIONS_PATH is not set in settings.py"):
            load_static_role_permissions()

    @override_settings(STATIC_ROLE_PERMISSIONS_PATH="tests.static_role_permissions.dummy_permissions.PERMISSIONS_LIST")
    def test_load_permissions_from_settings_path_with_invalid_dict(self):
        with self.assertRaisesMessage(AssertionError, "STATIC_ROLE_PERMISSIONS_PATH must point to a dict"):
            load_static_role_permissions()

    def test_has_perm_without_user_role(self):
        user = User()  # user.role does not exist
        self.assertEqual(False, user.has_perm("auth.add_user"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={
            "role_1": {"app_label.permission_1", "app_label.permission_2"},
        },
    )
    def test_has_perm_true_with_role_in_dict(self, _):
        user = User()
        user.role = "role_1"
        self.assertEqual(True, user.has_perm("app_label.permission_1"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={
            "role_1": {"app_label.permission_1", "app_label.permission_2"},
        },
    )
    def test_has_perm_false_with_perm_not_in_set(self, _):
        user = User()
        user.role = "role_1"
        self.assertEqual(False, user.has_perm("app_label.permission_3"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={
            "role_1": {},
        },
    )
    def test_has_perm_false_with_empty_set(self, _):
        user = User()
        user.role = "role_1"
        self.assertEqual(False, user.has_perm("app_label.permission_1"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={
            "role_2": {"app_label.permission_1"},  # <-- role_1 is missing
        },
    )
    def test_has_perm_false_with_role_not_in_dict(self, _):
        user = User()
        user.role = "role_1"
        self.assertEqual(False, user.has_perm("app_label.permission_1"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={
            "role_1": {"app_label.permission_1"},
        },
    )
    def test_has_perm_false_for_non_active_user(self, _):
        user = User()
        user.role = "role_1"
        user.is_active = False
        self.assertEqual(False, user.has_perm("app_label.permission_1"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={},
    )
    def test_has_perm_false_for_superuser_without_role(self, _):
        user = User()
        user.is_superuser = True
        self.assertEqual(True, user.has_perm("foo"))

    @mock.patch(
        "ambient_toolbox.static_role_permissions.auth_backend.load_static_role_permissions",
        return_value={
            "role_1": {"app_label.permission_1"},
        },
    )
    def test_has_perm_false_for_anonymous_user(self, _):
        anon = AnonymousUser()
        anon.role = "role_1"
        self.assertEqual(False, anon.has_perm("app_label.permission_1"))



import logging
from typing import ClassVar, Final

from django.conf import settings
from django.contrib.admin.models import LogEntry
from django.contrib.auth.hashers import make_password
from django.contrib.sessions.models import Session
from django.core.management import call_command
from django.db import OperationalError, connections


class ScrubbingError(RuntimeError):
    pass


class AbstractScrubbingService:
    DEFAULT_USER_PASSWORD: Final[str] = "Admin0404!"

    # Over-writable values
    keep_session_data: bool = False
    keep_scrubber_data: bool = False
    keep_django_admin_log: bool = False
    pre_scrub_functions: ClassVar = []
    post_scrub_functions: ClassVar = []

    def __init__(self):
        self._logger = logging.getLogger("django_scrubber")

    def _get_hashed_default_password(self):
        return make_password(self.DEFAULT_USER_PASSWORD)

    def _validation(self):
        if not settings.DEBUG:
            self._logger.warning("Attention! Has to run in DEBUG mode!")
            return False

        if "django_scrubber" not in settings.INSTALLED_APPS:
            self._logger.warning("Attention! django-scrubber needs to be installed!")
            return False

        if "django_scrubber" not in list(settings.LOGGING["loggers"].keys()):
            self._logger.warning("Attention! Logging for django-scrubber is not activated!")

        return True

    def process(self):
        self._logger.info("Start scrubbing process...")

        self._logger.info("Validating setup...")
        if not self._validation():
            self._logger.warning("Aborting process!")
            raise ScrubbingError("Scrubber settings validation failed")

        # Custom pre-scrubbing
        for name in self.pre_scrub_functions:
            self._logger.info(f'Pre-Scrubbing: Calling "{name}()"...')
            method = getattr(self, name)
            method()

        self._logger.info('Scrubbing data with "scrub_data"...')
        call_command("scrub_data")

        # Custom post-scrubbing
        for name in self.post_scrub_functions:
            self._logger.info(f'Post-Scrubbing: Calling "{name}()"...')
            method = getattr(self, name)
            method()

        # Empty django admin log (may contain user-related data)
        if not self.keep_django_admin_log:
            LogEntry.objects.all().delete()

        # Clear django session table
        if not self.keep_session_data:
            self._logger.info('Clearing data from "django_session" ...')
            # Sessions might contain private information and furthermore cannot be used anyway because we anonymised
            # all the users. Therefore, it is being cleared by default
            Session.objects.all().delete()

        # Reset scrubber data to avoid huge db dumps
        if not self.keep_scrubber_data:
            self._logger.info('Clearing data from "django_scrubber_fakedata" ...')
            # We truncate and don't scrub because the table is huge, and clearing on object-level might take a while.
            # Furthermore, we can avoid having a direct dependency on django-scrubber this way.
            cursor = connections["default"].cursor()
            try:
                cursor.execute("TRUNCATE TABLE django_scrubber_fakedata;")
            except OperationalError:
                # SQLite doesn't have "TRUNCATE"
                cursor.execute("DELETE FROM django_scrubber_fakedata;")

        self._logger.info("Scrubbing finished!")

        return True



from django import template

register = template.Library()


@register.filter
def format_to_minutes(time):
    """
    Converts the seconds to minutes

    :param time:
    :return minutes:
    """
    return time.seconds // 60



from contextvars import ContextVar
from typing import TYPE_CHECKING, Callable, Optional

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse

_request_cv: ContextVar[Optional["HttpRequest"]] = ContextVar("request", default=None)


class CurrentRequestMiddleware:
    """
    Middleware which stores the current request in a thread-safe manner.
    """

    def __init__(self, get_response: Callable[["HttpRequest"], "HttpResponse"]):
        self.get_response = get_response

    def __call__(self, request: "HttpRequest") -> "HttpResponse":
        token = _request_cv.set(request)
        response = self.get_response(request)
        _request_cv.reset(token)
        return response

    @staticmethod
    def get_current_user():
        request = _request_cv.get()
        try:
            return request.user
        except AttributeError:
            return None



class TempDisconnectSignal:
    """
    Context manager to temporarily disconnect a model from a signal.
    Attention: If your signal has a `dispatch_uid` set, you need to pass it to this class!
    Use with a "with" tag like this:
    ```
    with TempDisconnectSignal(**kwargs):
        obj.save()
    ```
    """

    def __init__(self, signal, receiver, sender, dispatch_uid=None):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender
        self.dispatch_uid = dispatch_uid

    def __enter__(self):
        self.signal.disconnect(
            receiver=self.receiver,
            sender=self.sender,
            dispatch_uid=self.dispatch_uid,
        )

    def __exit__(self, type, value, traceback):  # noqa: A002
        self.signal.connect(
            receiver=self.receiver,
            sender=self.sender,
            dispatch_uid=self.dispatch_uid,
        )



from django.contrib.auth.models import AbstractUser, User
from django.test import TestCase
from rest_framework import status
from rest_framework.reverse import reverse

from ambient_toolbox.drf.tests import BaseViewSetTestMixin
from testapp.api import views
from testapp.models import MySingleSignalModel


class BaseApiTest(BaseViewSetTestMixin, TestCase):
    def get_default_api_user(self) -> AbstractUser:
        return User.objects.create(username="my-username", is_active=True)


class MySingleSignalModelApiViewTest(BaseApiTest):
    view_class = views.MySingleSignalModelViewSet

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        # Create list of objects
        cls.object_list = [
            MySingleSignalModel.objects.create(value=1),
            MySingleSignalModel.objects.create(value=2),
        ]

    def test_action_not_activated(self):
        with self.assertRaises(AttributeError):
            self.execute_request(
                method="post",
                url=reverse("my-single-signal-model-list"),
                viewset_kwargs={"post": "create"},
                user=self.default_api_user,
            )

    def test_list_authentication_required(self):
        self.validate_authentication_required(url=reverse("my-single-signal-model-list"), method="get", view="list")

    def test_list_regular(self):
        response = self.execute_request(
            method="get",
            url=reverse("my-single-signal-model-list"),
            viewset_kwargs={"get": "list"},
            user=self.default_api_user,
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), len(self.object_list))



# Generated by Django 5.0.7 on 2024-07-15 14:19

import ambient_toolbox.mixins.bleacher
import ambient_toolbox.mixins.models
import ambient_toolbox.mixins.validation
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="MyPermissionModelMixin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "managed": False,
            },
            bases=(ambient_toolbox.mixins.models.PermissionModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="BleacherMixinModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("content", models.CharField(max_length=50)),
            ],
            bases=(ambient_toolbox.mixins.bleacher.BleacherMixin, models.Model),
        ),
        migrations.CreateModel(
            name="ModelWithCleanMixin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            bases=(ambient_toolbox.mixins.validation.CleanOnSaveMixin, models.Model),
        ),
        migrations.CreateModel(
            name="ModelWithSaveWithoutSignalsMixin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.PositiveIntegerField(default=0)),
            ],
            bases=(ambient_toolbox.mixins.models.SaveWithoutSignalsMixin, models.Model),
        ),
        migrations.CreateModel(
            name="ModelWithSelector",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name="MyMultipleSignalModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name="MySingleSignalModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.PositiveIntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name="CommonInfoBasedModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        verbose_name="Created at",
                    ),
                ),
                (
                    "lastmodified_at",
                    models.DateTimeField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        verbose_name="Last modified at",
                    ),
                ),
                ("value", models.PositiveIntegerField(default=0)),
                ("value_b", models.PositiveIntegerField(default=0)),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(app_label)s_%(class)s_created",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Created by",
                    ),
                ),
                (
                    "lastmodified_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="%(app_label)s_%(class)s_lastmodified",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Last modified by",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ModelWithFkToSelf",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="testapp.modelwithfktoself",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ModelWithOneToOneToSelf",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "peer",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="related_peer",
                        to="testapp.modelwithonetoonetoself",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ForeignKeyRelatedModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "single_signal",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="foreign_key_related_models",
                        to="testapp.mysinglesignalmodel",
                    ),
                ),
            ],
        ),
    ]



from django.test import TestCase

from ambient_toolbox.utils.math import round_to_decimal, round_up_decimal


class MathTest(TestCase):
    def test_round_to_decimal_no_precision(self):
        self.assertEqual(round_to_decimal(5.2, 1), 5)

    def test_round_to_decimal_two_precisions(self):
        self.assertEqual(round_to_decimal(5.62, 0.05), 5.60)

    def test_round_to_decimal_down_lower(self):
        self.assertEqual(round_to_decimal(5.2), 5.0)

    def test_round_to_decimal_up_lower(self):
        self.assertEqual(round_to_decimal(5.4), 5.5)

    def test_round_to_decimal_down_upper(self):
        self.assertEqual(round_to_decimal(5.6), 5.5)

    def test_round_to_decimal_no_round(self):
        self.assertEqual(round_to_decimal(5.0), 5.0)

    def test_round_to_decimal_up_upper(self):
        self.assertEqual(round_to_decimal(5.8), 6.0)

    def test_round_up_decimal_no_precision(self):
        self.assertEqual(round_up_decimal(5.2, 1), 6)

    def test_round_up_decimal_two_precisions(self):
        self.assertEqual(round_up_decimal(5.62, 0.05), 5.65)

    def test_round_up_decimal_down_lower(self):
        self.assertEqual(round_up_decimal(5.2), 5.5)

    def test_round_up_decimal_up_lower(self):
        self.assertEqual(round_up_decimal(5.4), 5.5)

    def test_round_up_decimal_down_upper(self):
        self.assertEqual(round_up_decimal(5.6), 6.0)

    def test_round_up_decimal_no_round(self):
        self.assertEqual(round_up_decimal(5.0), 5.0)

    def test_round_up_decimal_up_upper(self):
        self.assertEqual(round_up_decimal(5.8), 6.0)



import pytest

from ambient_toolbox.utils.file import crc, get_filename_without_ending, md5_checksum


def test_get_filename_without_ending_full_path():
    assert get_filename_without_ending("path/to/my/text-file.txt") == "text-file"


def test_get_filename_without_ending_only_filename():
    assert get_filename_without_ending("text-file.txt") == "text-file"


@pytest.fixture
def gen_test_file(tmp_path):
    def inner(content):
        test_file = tmp_path / "test_file.txt"
        test_file.write_text(content)
        return test_file

    return inner


@pytest.mark.parametrize("test_func", [crc, md5_checksum])
def test_closes_file(mocker, test_func):
    """
    Tests if the CRC and MD5 checksum functions use a context manager to open the file, to guarantee that the opened
    file descriptor is closed.
    """
    open_mock = mocker.patch("ambient_toolbox.utils.file.open")
    open_mock.return_value.__enter__.return_value.read.return_value = None  # to make f.read() return None.
    file_mock = mocker.Mock()
    test_func(file_mock)
    assert open_mock.call_args[0][0] == file_mock
    assert open_mock.return_value.__exit__.call_count == 1


@pytest.mark.parametrize(
    "content, crc_result, md5_result",
    [
        ("The answer to life, the universe, and everything.", "31F49620", "f81ab2f7fb6cacf50f973b0dc8faff44"),
        ("The quick brown fox jumps over the lazy dog", "414FA339", "9e107d9d372bb6826bd81d3542a419d6"),
        ("", "00000000", "d41d8cd98f00b204e9800998ecf8427e"),
    ],
)
def test_crc_and_md5(gen_test_file, content, crc_result, md5_result):
    """Generates the CRC and MD5 of a test file and checks its result"""
    test_file = str(gen_test_file(content))
    result = crc(test_file)
    assert result == crc_result
    result = md5_checksum(test_file)
    assert result == md5_result



from unittest import mock

from django.test import TestCase, override_settings

from ambient_toolbox.services.custom_scrubber import AbstractScrubbingService, ScrubbingError


class AbstractScrubbingServiceTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.service = AbstractScrubbingService()

    @override_settings(DEBUG=False)
    def test_scrubber_debug_mode_needs_to_be_active(self):
        with self.assertRaisesMessage(ScrubbingError, "Scrubber settings validation failed"):
            self.service.process()

    @override_settings(DEBUG=True, INSTALLED_APPS=[])
    def test_scrubber_needs_to_be_installed(self):
        with self.assertRaisesMessage(ScrubbingError, "Scrubber settings validation failed"):
            self.service.process()

    @mock.patch("ambient_toolbox.services.custom_scrubber.make_password")
    def test_get_hashed_default_password_regular(self, mocked_make_password):
        self.service._get_hashed_default_password()
        mocked_make_password.assert_called_once_with(self.service.DEFAULT_USER_PASSWORD)

    # TODO: write more tests



from django.core.management.base import BaseCommand

from ambient_toolbox.tests.structure_validator.test_structure_validator import TestStructureValidator


class Command(BaseCommand):
    def handle(self, *args, **options):
        service = TestStructureValidator()
        service.process()



SENTRY_EVENT = {
    "level": "error",
    "exception": {
        "values": [
            {
                "module": None,
                "type": "ZeroDivisionError",
                "value": "division by zero",
                "mechanism": {"type": "django", "handled": False},
                "stacktrace": {
                    "frames": [
                        {
                            "filename": "test/account/api/views.py",
                            "abs_path": "/opt/project/backend/test/account/api/views.py",
                            "function": "perform_update",
                            "module": "test.account.api.views",
                            "lineno": 123,
                            "pre_context": ["test = 1/0"],
                            "vars": {
                                "self": "<testapp.account.api.views.TestViewSet object>",
                                "serializer": "TestDetailSerializer(<User: Test AAAA>)",
                                "__class__": "<class 'test.account.api.views.TestViewSet'>",
                                "admin": "<EmailUser: Test User>",
                                "test_data": "123",
                            },
                            "in_app": True,
                        }
                    ]
                },
            }
        ]
    },
    "request": {
        "url": "http://localhost:8000/api/v2/test/17/",
        "query_string": "",
        "method": "PUT",
        "env": {"SERVER_NAME": "cdc3cb5b00af", "SERVER_PORT": "8000", "REMOTE_ADDR": "172.22.0.1"},
        "headers": {
            "Content-Length": "1202",
            "Content-Type": "application/json",
            "Host": "localhost:8000",
            "User-Agent": "Mozilla/5.0 (Macintosh;) Gecko/20100101 Firefox/112.0",
            "Accept": "application/json, text/plain, */*",
            "Accept-Language": "en",
            "Accept-Encoding": "gzip, deflate, br",
            "Authorization": "Token 123123123123123123123123123123123123abcd",
            "Origin": "http://127.0.0.1:3000",
            "Dnt": "1",
            "Connection": "keep-alive",
            "Referer": "http://127.0.0.1:3000/",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "cross-site",
        },
        "cookies": {},
        "data": "",
    },
    "user": {"email": "test@test.local", "id": "1337", "ip_address": "172.22.0.1", "username": "test@test.local"},
}

SCRUBBED_SENTRY_EVENT = {
    "level": "error",
    "exception": {
        "values": [
            {
                "module": None,
                "type": "ZeroDivisionError",
                "value": "division by zero",
                "mechanism": {"type": "django", "handled": False},
                "stacktrace": {
                    "frames": [
                        {
                            "filename": "test/account/api/views.py",
                            "abs_path": "/opt/project/backend/test/account/api/views.py",
                            "function": "perform_update",
                            "module": "test.account.api.views",
                            "lineno": 123,
                            "pre_context": ["test = 1/0"],
                            "vars": {
                                "self": "'<testapp.account.api.views.TestViewSet object>'",
                                "serializer": "[Filtered]",  # filtered due to standard filter
                                "__class__": "\"<class 'test.account.api.views.TestViewSet'>\"",
                                "admin": "[Filtered]",  # filtered due to standard filter
                                "test_data": "'123'",
                            },
                            "in_app": True,
                        }
                    ]
                },
            }
        ]
    },
    "request": {
        "url": "http://localhost:8000/api/v2/test/17/",
        "query_string": "",
        "method": "PUT",
        "env": {"SERVER_NAME": "cdc3cb5b00af", "SERVER_PORT": "8000", "REMOTE_ADDR": "172.22.0.1"},
        "headers": {
            "Content-Length": "1202",
            "Content-Type": "application/json",
            "Host": "localhost:8000",
            "User-Agent": "Mozilla/5.0 (Macintosh;) Gecko/20100101 Firefox/112.0",
            "Accept": "application/json, text/plain, */*",
            "Accept-Language": "en",
            "Accept-Encoding": "gzip, deflate, br",
            "Authorization": "[Filtered]",  # filtered due to default sentry filter
            "Origin": "http://127.0.0.1:3000",
            "Dnt": "1",
            "Connection": "keep-alive",
            "Referer": "http://127.0.0.1:3000/",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "cross-site",
        },
        "cookies": {},
        "data": "",
    },
    "user": {
        "email": "[Filtered]",  # filtered due to default sentry filter
        "id": "1337",
        "ip_address": "[Filtered]",  # filtered due to default sentry filter
        "username": "[Filtered]",  # filtered due to default sentry filter
    },
    "_meta": {
        "exception": {
            "values": {
                "0": {
                    "stacktrace": {
                        "frames": {
                            "0": {
                                "vars": {
                                    "serializer": {"": {"rem": [["!config", "s"]]}},
                                    "admin": {"": {"rem": [["!config", "s"]]}},
                                }
                            }
                        }
                    }
                }
            }
        },
        "request": {"headers": {"Authorization": {"": {"rem": [["!config", "s"]]}}}},
        "user": {
            "email": {"": {"rem": [["!config", "s"]]}},
            "ip_address": {"": {"rem": [["!config", "s"]]}},
            "username": {"": {"rem": [["!config", "s"]]}},
        },
    },
}



def generate_default_permissions(model_name: str) -> list[str]:
    return [
        f"add_{model_name}",
        f"change_{model_name}",
        f"delete_{model_name}",
        f"view_{model_name}",
    ]



import copy

from django.test import TestCase

from ambient_toolbox.sentry.helpers import SentryEventScrubber, strip_sensitive_data_from_sentry_event
from tests.sentry.mock_data import SCRUBBED_SENTRY_EVENT, SENTRY_EVENT


class SentryHelperTest(TestCase):
    def test_strip_sensitive_data_from_sentry_event_regular(self):
        event = {"user": {"email": "mymail@example.com", "ip_address": "127.0.0.1", "username": "my-user"}}

        self.assertIsInstance(strip_sensitive_data_from_sentry_event(event, None), dict)

    def test_strip_sensitive_data_from_sentry_event_missing_key_email(self):
        event = {"user": {"ip_address": "127.0.0.1", "username": "my-user"}}

        self.assertIsInstance(strip_sensitive_data_from_sentry_event(event, None), dict)

    def test_strip_sensitive_data_from_sentry_event_missing_key_ip_address(self):
        event = {"user": {"email": "mymail@example.com", "username": "my-user"}}

        self.assertIsInstance(strip_sensitive_data_from_sentry_event(event, None), dict)

    def test_strip_sensitive_data_from_sentry_event_missing_key_username(self):
        event = {"user": {"email": "mymail@example.com", "ip_address": "127.0.0.1"}}

        self.assertIsInstance(strip_sensitive_data_from_sentry_event(event, None), dict)

    def test_sentry_event_scrubber_on_a_regular_sentry_event(self):
        event = copy.deepcopy(SENTRY_EVENT)
        self.assertEqual(
            SCRUBBED_SENTRY_EVENT,
            SentryEventScrubber().scrub_sensitive_data_from_sentry_event(event, None),
        )



import json
from typing import Optional

from django.test import Client, TestCase


class GraphQLTestCase(TestCase):
    """
    Provides a best-practice wrapper for easily testing GraphQL endpoints.
    """

    # URL to graphql endpoint
    GRAPHQL_URL = "/graphql/"
    # Here you need to set your graphql schema for the tests
    GRAPHQL_SCHEMA = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        if not cls.GRAPHQL_SCHEMA:
            raise AttributeError("Variable GRAPHQL_SCHEMA not defined in GraphQLTestCase.")

        cls._client = Client(cls.GRAPHQL_SCHEMA)

    def query(self, query: str, op_name: Optional[str] = None, input_data: Optional[dict] = None):
        """
        :param query:       GraphQL query to run
        :param op_name:     If the query is a mutation or named query, you must supply the op_name.
                            For annon queries ("{ ... }"), should be None (default).
        :param input_data:  If provided, the $input variable in GraphQL will be set to this value
        :return:            Response object from client
        """
        body = {"query": query}
        if op_name:
            body["operation_name"] = op_name
        if input_data:
            body["variables"] = {"input": input_data}

        resp = self._client.post(self.GRAPHQL_URL, json.dumps(body), content_type="application/json")
        return resp

    def assertResponseNoErrors(self, resp):  # noqa: N802
        """
        Assert that the call went through correctly. 200 means the syntax is ok, if there are no `errors`,
        the call was fine.
        :resp HttpResponse: Response
        """
        content = json.loads(resp.content)
        self.assertEqual(resp.status_code, 200)
        self.assertNotIn("errors", list(content.keys()))

    def assertResponseHasErrors(self, resp):  # noqa: N802
        """
        Assert that the call was failing. Take care: Even with errors, GraphQL returns status 200!
        :resp HttpResponse: Response
        """
        content = json.loads(resp.content)
        self.assertIn("errors", list(content.keys()))



from django.conf import settings
from django.utils.module_loading import import_string


def load_static_role_permissions() -> dict[str, set[str]]:
    """
    Load static definition of permissions by role.

    Example:
    {
        "role_1": {
            "app_label.permission",
            "app_label.permission",
            ...
        },
        "role_2": {
            "app_label.permission",
            "app_label.permission",
            ...
        },
    """
    dotted_path = getattr(settings, "STATIC_ROLE_PERMISSIONS_PATH", None)
    assert dotted_path, "STATIC_ROLE_PERMISSIONS_PATH is not set in settings.py"

    permission = import_string(dotted_path)
    assert isinstance(permission, dict), "STATIC_ROLE_PERMISSIONS_PATH must point to a dict"

    return permission



from django.contrib.auth.backends import ModelBackend

from ambient_toolbox.static_role_permissions.permissions import load_static_role_permissions


class StaticRolePermissionBackend(ModelBackend):
    """
    This ModelBackend relies on a static definition of permissions,
    instead of pulling them from the database.

    It expects the user model to have a `role` attribute,
    wich matches with the keys in the STATIC_ROLE_PERMISSIONS dict.

    Example:

    STATIC_ROLE_PERMISSIONS = {
        "admin": {
            "auth.add_user",
            "auth.change_user",
            "auth.delete_user",
        },
        "editor": {
            "auth.add_user",
            "auth.change_user",
        },
        "viewer": {
            "auth.view_user",
        },
    }

    user.role = "admin"

    user.has_perm("auth.add_user") -> True
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._static_permissions_cache = None

    def _get_static_role_permissions(self):
        if self._static_permissions_cache is None:
            self._static_permissions_cache = load_static_role_permissions()
        return self._static_permissions_cache

    def _get_permissions_for_role(self, role: str) -> set[str]:
        role_permissions_dict = self._get_static_role_permissions()
        return role_permissions_dict.get(role, set())

    def get_all_permissions(self, user_obj, obj=None):
        if not user_obj.is_active or user_obj.is_anonymous or obj is not None:
            return set()
        role = getattr(user_obj, "role", None)

        if role is None:
            return set()

        return self._get_permissions_for_role(role)



from typing import Optional

from django.db.models import ForeignKey, Model


def object_to_dict(obj, blacklisted_fields: Optional[list] = None, include_id: bool = False) -> dict:
    """
    Returns a dict with all data defined in the model class as a key-value-dict
    Attention: Does not work for M2M fields!
    """
    # Default blacklist
    blacklisted_fields = blacklisted_fields if blacklisted_fields else []

    # Add default django primary key to blacklist
    if not include_id:
        blacklisted_fields.append("id")

    data = vars(obj)

    valid_fields = []
    for f in obj.__class__._meta.get_fields():
        if not isinstance(f, ForeignKey):
            valid_fields.append(f.name)
        else:
            valid_fields.append(f"{f.name}_id")

    valid_data = {key: value for key, value in data.items() if key in valid_fields and key not in blacklisted_fields}

    return valid_data


def get_cached_related_obj(obj: Model, related_field_name: str, silently_return_none: bool = False):
    """
    This function helps to avoid silent sub-queries, due to missing `select_related()` or `prefetch_related()`.

    This function performs a lookup in the fields_cache of the given object.
    If silently_return_none is True, this function will return None if the field is not cached.
    Otherwise, an AttributeError will be raised.

    What problem does this function solve?

    foo = Foo.objects.all().get()
    foo.bar  # <-- this will silently cause a sub-query

    foo = Foo.objects.all().select_related('bar').first()
    foo.bar # <-- this will cause no sub-query

    The only difference is the usage of `select_related()`, which can easily be forgotten.

    Usage:
    get_cached_related_obj(foo, 'bar')

    This will result in a KeyError instead of a silent sub-query if the field is not cached
    via `select_related()` or `prefetch_related()`.
    """
    # warning: this is an undocumented feature of django
    fields_cache = obj._state.fields_cache

    if related_field_name in fields_cache:
        return fields_cache[related_field_name]

    if silently_return_none:
        return None

    raise AttributeError(
        f'Field "{related_field_name}" not found in `fields_cache` of {obj.__class__.__name__} object. '
        "Did you forget to use `select_related()` or `prefetch_related()`?"
    )



from typing import ClassVar

from django.contrib.auth.models import AnonymousUser
from django.test import TestCase
from django.views import generic

from ambient_toolbox.tests.mixins import RequestProviderMixin
from ambient_toolbox.view_layer.htmx_mixins import HtmxResponseMixin


class HtmxResponseMixinTest(RequestProviderMixin, TestCase):
    class TestView(HtmxResponseMixin, generic.View):
        hx_redirect_url = "https://my-url.com"
        hx_trigger = "myEvent"

    class TestViewWithTriggerDict(HtmxResponseMixin, generic.View):
        hx_trigger: ClassVar = {"myEvent": None}

    def test_dispatch_functional(self):
        view = self.TestView()

        response = view.dispatch(request=self.get_request(user=AnonymousUser()))

        self.assertIn("HX-Redirect", response)
        self.assertEqual(response["HX-Redirect"], "https://my-url.com")

        self.assertIn("HX-Trigger", response)
        self.assertEqual(response["HX-Trigger"], "myEvent")

    def test_dispatch_trigger_with_dict(self):
        view = self.TestViewWithTriggerDict()

        response = view.dispatch(request=self.get_request(user=AnonymousUser()))

        self.assertIn("HX-Trigger", response)
        self.assertEqual(response["HX-Trigger"], '{"myEvent": null}')

    def test_get_hx_redirect_url_regular(self):
        view = self.TestView()

        self.assertEqual(view.get_hx_redirect_url(), "https://my-url.com")

    def test_get_hx_trigger_regular(self):
        view = self.TestView()

        self.assertEqual(view.get_hx_trigger(), "myEvent")



from pathlib import Path
from unittest import mock

from django.conf import settings
from django.test import TestCase, override_settings

from ambient_toolbox.tests.structure_validator.test_structure_validator import TestStructureValidator


class TestStructureValidatorTest(TestCase):
    def test_init_regular(self):
        service = TestStructureValidator()

        self.assertEqual(service.file_whitelist, ["__init__"])
        self.assertEqual(service.issue_list, [])

    @override_settings(TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST=["my_file"])
    def test_get_file_whitelist_from_settings(self):
        service = TestStructureValidator()
        file_whitelist = service._get_file_whitelist()

        self.assertEqual(file_whitelist, ["__init__", "my_file"])

    def test_get_file_whitelist_fallback(self):
        service = TestStructureValidator()
        file_whitelist = service._get_file_whitelist()

        self.assertEqual(file_whitelist, ["__init__"])

    @override_settings(TEST_STRUCTURE_VALIDATOR_BASE_DIR=settings.BASE_PATH)
    def test_get_base_dir_from_settings(self):
        service = TestStructureValidator()
        base_dir = service._get_base_dir()

        self.assertEqual(base_dir, settings.BASE_PATH)

    def test_get_base_dir_fallback(self):
        service = TestStructureValidator()
        base_dir = service._get_base_dir()

        self.assertEqual(base_dir, "")

    @override_settings(TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME="my_project")
    def test_get_base_app_name_from_settings(self):
        service = TestStructureValidator()
        base_app_name = service._get_base_app_name()

        self.assertEqual(base_app_name, "my_project")

    def test_get_base_app_name_fallback(self):
        service = TestStructureValidator()
        base_app_name = service._get_base_app_name()

        self.assertEqual(base_app_name, "apps")

    @override_settings(TEST_STRUCTURE_VALIDATOR_APP_LIST=["apps.my_app", "apps.other_app"])
    def test_get_app_list_from_settings(self):
        service = TestStructureValidator()
        base_app_name = service._get_app_list()

        self.assertEqual(base_app_name, ["apps.my_app", "apps.other_app"])

    def test_get_app_list_fallback(self):
        service = TestStructureValidator()
        base_app_name = service._get_app_list()

        self.assertEqual(base_app_name, settings.INSTALLED_APPS)

    @override_settings(TEST_STRUCTURE_VALIDATOR_IGNORED_DIRECTORY_LIST=["my_dir", "other_dir"])
    def test_get_ignored_directory_list_from_settings(self):
        service = TestStructureValidator()
        dir_list = service._get_ignored_directory_list()

        self.assertEqual(dir_list, ["__pycache__", "my_dir", "other_dir"])

    def test_get_ignored_directory_list_fallback(self):
        service = TestStructureValidator()
        dir_list = service._get_ignored_directory_list()

        self.assertEqual(dir_list, ["__pycache__"])

    def test_check_missing_test_prefix_correct_prefix(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root="root/path",
            file="missing_prefix",
            filename="test_my_file",
            extension=".py",
        )

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    @override_settings(TEST_STRUCTURE_VALIDATOR_FILE_WHITELIST=["my_file"])
    def test_check_missing_test_prefix_wrong_prefix_but_whitelisted(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root="root/path",
            file="missing_prefix",
            filename="my_file",
            extension=".py",
        )

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_test_prefix_wrong_prefix_but_not_py_file(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root="root/path",
            file="missing_prefix",
            filename="missing_prefix",
            extension=".txt",
        )

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_test_prefix_wrong_prefix(self):
        service = TestStructureValidator()
        result = service._check_missing_test_prefix(
            root="root/path",
            file="missing_prefix",
            filename="missing_prefix",
            extension=".py",
        )

        self.assertFalse(result)
        self.assertEqual(len(service.issue_list), 1)

    def test_check_missing_init_init_found_files_in_dir(self):
        service = TestStructureValidator()
        result = service._check_missing_init(root="root/path", init_found=True, number_of_py_files=1)

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_init_no_init_no_files(self):
        service = TestStructureValidator()
        result = service._check_missing_init(root="root/path", init_found=False, number_of_py_files=0)

        self.assertTrue(result)
        self.assertEqual(len(service.issue_list), 0)

    def test_check_missing_init_no_init_but_files(self):
        service = TestStructureValidator()
        result = service._check_missing_init(root="root/path", init_found=False, number_of_py_files=1)

        self.assertFalse(result)
        self.assertEqual(len(service.issue_list), 1)

    @override_settings(
        TEST_STRUCTURE_VALIDATOR_BASE_DIR=Path("/src/ambient_toolbox/"),
        TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME="my_project",
    )
    def test_build_path_to_test_package_with_settings_path(self):
        service = TestStructureValidator()
        path = service._build_path_to_test_package(app="my_project.my_app")

        self.assertEqual(path, Path("/src/ambient_toolbox/my_project/my_app/tests"))

    @override_settings(
        TEST_STRUCTURE_VALIDATOR_BASE_DIR="/src/ambient_toolbox/", TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME="my_project"
    )
    def test_build_path_to_test_package_with_settings_str(self):
        service = TestStructureValidator()
        path = service._build_path_to_test_package(app="my_project.my_app")

        self.assertEqual(path, Path("/src/ambient_toolbox/my_project/my_app/tests"))

    def test_build_path_to_test_package_with_defaults(self):
        service = TestStructureValidator()
        path = service._build_path_to_test_package(app="my_project.my_app")

        self.assertEqual(path, Path("my_project/my_app/tests"))

    @override_settings(
        TEST_STRUCTURE_VALIDATOR_BASE_DIR=settings.BASE_PATH,
        TEST_STRUCTURE_VALIDATOR_APP_LIST=["testapp"],
        TEST_STRUCTURE_VALIDATOR_BASE_APP_NAME="",
    )
    def test_process_functional(self):
        service = TestStructureValidator()
        with self.assertRaises(SystemExit):
            service.process()

        self.assertEqual(len(service.issue_list), 2)

        complaint_list = sorted(service.issue_list)

        self.assertIn('Python file without "test_" prefix found:', complaint_list[0])
        self.assertIn("testapp/tests/subdirectory/missing_test_prefix.py", complaint_list[0])

        self.assertIn("__init__.py missing in", complaint_list[1])
        self.assertIn("testapp/tests/missing_init", complaint_list[1])

    @mock.patch.object(TestStructureValidator, "_get_app_list", return_value=["invalidly_located_app"])
    def test_process_invalidly_located_app(self, mocked_get_app_list):
        service = TestStructureValidator()

        service.process()

        mocked_get_app_list.assert_called_once()



from django.db import models


def log_whodid(obj: models.Model, user) -> None:
    """
    Stores the given user as creator or editor of the given object
    """
    if hasattr(obj, "created_by") and obj.created_by is None:
        obj.created_by = user

    if hasattr(obj, "lastmodified_by"):
        obj.lastmodified_by = user



from unittest import mock

from django.test import TestCase

from ambient_toolbox.mixins.bleacher import BleacherMixin
from testapp.models import BleacherMixinModel


class BleacherMixinTest(TestCase):
    @mock.patch.object(BleacherMixin, "DEFAULT_ALLOWED_TAGS", ["a", "b", "p"])
    def test_init_allowed_tags_casted_to_set(self, *args):
        obj = BleacherMixinModel()
        self.assertEqual({"a", "b", "p"}, obj.allowed_tags)
        self.assertIs(True, isinstance(obj.allowed_tags, set))

    @mock.patch.object(BleacherMixin, "DEFAULT_ALLOWED_ATTRIBUTES", {"img": ("alt",)})
    def test_init_allowed_attributes_casted_to_set(self):
        obj = BleacherMixinModel()
        self.assertEqual({"img": {"alt"}}, obj.allowed_attributes)
        self.assertIs(True, isinstance(obj.allowed_attributes["img"], set))

    def test_save_allowed_tag(self):
        obj = BleacherMixinModel.objects.create(content="<p>Test</p>")
        self.assertEqual("<p>Test</p>", obj.content)

    def test_save_remove_tags_with_adjacent_content(self):
        obj = BleacherMixinModel.objects.create(content="<script>Evil</script>Test")
        self.assertEqual("Test", obj.content)

    def test_save_default_rel_behaviour(self):
        obj = BleacherMixinModel.objects.create(content="<a rel='stuff'>Test</a>")
        self.assertEqual('<a rel="noopener noreferrer">Test</a>', obj.content)

    def test_save_tag_removed_content_stays(self):
        obj = BleacherMixinModel.objects.create(content="<modal>Test</modal>")
        self.assertEqual("Test", obj.content)

    @mock.patch.object(BleacherMixinModel, "DEFAULT_ALLOWED_TAGS", ["modal"])
    def test_save_custom_tag_allowed(self):
        obj = BleacherMixinModel.objects.create(content="<modal>Test</modal>")
        self.assertEqual("<modal>Test</modal>", obj.content)

    @mock.patch.object(BleacherMixinModel, "DEFAULT_ALLOWED_ATTRIBUTES", {"img": ("alt",)})
    def test_save_custom_attribute_allowed(self):
        obj = BleacherMixinModel.objects.create(content='<img alt="Noodle" />')
        self.assertEqual('<img alt="Noodle">', obj.content)



from django.contrib.auth.models import User
from django.test import TestCase

from ambient_toolbox.managers import (
    AbstractUserSpecificManager,
    AbstractUserSpecificQuerySet,
)
from testapp.models import MySingleSignalModel


class AbstractUserSpecificQuerySetTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.cqs = AbstractUserSpecificQuerySet()
        cls.user = User.objects.create(username="my-username")

    def test_default(self):
        self.assertEqual(self.cqs.default(self.user), self.cqs)

    def test_visible_for_regular(self):
        with self.assertRaisesMessage(NotImplementedError, "Please implement this method"):
            self.cqs.visible_for(self.user)

    def test_editable_for_regular(self):
        with self.assertRaisesMessage(NotImplementedError, "Please implement this method"):
            self.cqs.editable_for(self.user)

    def test_deletable_for_regular(self):
        with self.assertRaisesMessage(NotImplementedError, "Please implement this method"):
            self.cqs.deletable_for(self.user)


class TestUserSpecificManager(AbstractUserSpecificManager):
    pass


TestUserSpecificManager = TestUserSpecificManager.from_queryset(AbstractUserSpecificQuerySet)


class AbstractUserSpecificManagerTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.manager = TestUserSpecificManager()
        cls.user = User.objects.create(username="my-username")

    def test_visible_for_regular(self):
        with self.assertRaisesMessage(NotImplementedError, "Please implement this method"):
            self.manager.visible_for(self.user)

    def test_editable_for_regular(self):
        with self.assertRaisesMessage(NotImplementedError, "Please implement this method"):
            self.manager.editable_for(self.user)

    def test_deletable_for_regular(self):
        with self.assertRaisesMessage(NotImplementedError, "Please implement this method"):
            self.manager.deletable_for(self.user)


class GloballyVisibleQuerySetTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        # Create test user
        cls.user = User.objects.create(username="my-username")

        # Create list of objects
        cls.object_list = [
            MySingleSignalModel.objects.create(value=1),
            MySingleSignalModel.objects.create(value=2),
        ]

    def test_visible_for_regular(self):
        self.assertGreater(len(self.object_list), 0)
        self.assertEqual(
            MySingleSignalModel.objects.visible_for(self.user).count(),
            len(self.object_list),
        )

    def test_editable_for_regular(self):
        self.assertGreater(len(self.object_list), 0)
        self.assertEqual(
            MySingleSignalModel.objects.editable_for(self.user).count(),
            len(self.object_list),
        )

    def test_deletable_for_regular(self):
        self.assertGreater(len(self.object_list), 0)
        self.assertEqual(
            MySingleSignalModel.objects.deletable_for(self.user).count(),
            len(self.object_list),
        )



from django.contrib.auth.models import User
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _


class DjangoPermissionRequiredMixin:
    """
    View mixin to enforce logged-in state and Django permissions.
    """

    permission_list = None
    login_required = True
    login_view_name = "login-view"

    def __init__(self):
        super().__init__()

        if self.permission_list is None:
            raise RuntimeError(
                _("Class-based view using DjangoPermissionRequiredMixin without defining a permission list.")
            )

    def get_login_url(self) -> str:
        """
        Method that can be overwritten to define the login url. If a user has to be logged in but is not, he/she
        will be forwarded to the login view.
        """
        return reverse(self.login_view_name)

    def passes_login_barrier(self, user) -> bool:
        """
        If user does not have to be logged in, we just let anybody pass. Otherwise, user has to be logged in.
        """
        if not self.login_required or user.is_authenticated:
            return True

        return False

    def has_permissions(self, user: User) -> bool:
        # Check every permission...
        for permission in self.permission_list:
            # If user is missing one, the user is not allowed to see the view
            if not user.has_perm(permission):
                return False

        # If all permissions validate, the user can access the view
        return True

    def dispatch(self, request, *args, **kwargs):
        # Validate user is either logged in or doesn't have to be logged in
        if not self.passes_login_barrier(request.user):
            return redirect(self.get_login_url())

        # Validate that user has all required permissions
        if not self.has_permissions(request.user):
            return render(request, "403.html", status=403)

        # If everything goes well, we'll continue to the view
        return super().dispatch(request, *args, **kwargs)



from django import forms

from testapp.models import CommonInfoBasedModel


class CommonInfoBasedModelTestForm(forms.ModelForm):
    class Meta:
        model = CommonInfoBasedModel
        fields = ("value",)



import dataclasses


@dataclasses.dataclass
class PermissionModelDeclaration:
    app_label: str
    codename_list: list[str]
    model: str


@dataclasses.dataclass
class GroupPermissionDeclaration:
    name: str
    permission_list: list[PermissionModelDeclaration]



from django.views import generic

from ambient_toolbox.view_layer.views import UserInFormKwargsMixin
from testapp.forms import CommonInfoBasedModelTestForm


class UserInFormKwargsMixinView(UserInFormKwargsMixin, generic.FormView):
    """
    Generic view for updating an object without any user data being sent. Therefore, we don't need a form to validate
    user input.
    Most common use-case is toggling a flag inside an object.
    """

    form_class = CommonInfoBasedModelTestForm



from django.contrib import admin
from django.contrib.auth.models import User
from django.forms import forms
from django.test import TestCase

from ambient_toolbox.admin.model_admins.mixins import AdminCreateFormMixin
from ambient_toolbox.tests.mixins import RequestProviderMixin
from testapp.models import ForeignKeyRelatedModel, MySingleSignalModel


class TestCreateForm(forms.Form):
    class Meta:
        custom_add_form = True


class TestAdminCreateFormMixinAdmin(AdminCreateFormMixin, admin.ModelAdmin):
    add_form = TestCreateForm


class AdminCreateFormMixinTest(RequestProviderMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.super_user = User.objects.create(username="super_user", is_superuser=True)

        admin.site.register(ForeignKeyRelatedModel, TestAdminCreateFormMixinAdmin)

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        admin.site.unregister(ForeignKeyRelatedModel)

    def test_add_form_used_in_create_case(self):
        model_admin = TestAdminCreateFormMixinAdmin(model=ForeignKeyRelatedModel, admin_site=admin.site)

        form = model_admin.get_form(self.get_request(self.super_user))

        # Use Meta attribute from custom form class to determine if form was used. Base form is being wrapped
        # so we cannot assert the class.
        self.assertTrue(form.Meta.custom_add_form)

    def test_add_form_not_used_in_edit_case(self):
        model_admin = TestAdminCreateFormMixinAdmin(model=ForeignKeyRelatedModel, admin_site=admin.site)

        form = model_admin.get_form(
            self.get_request(self.super_user), obj=ForeignKeyRelatedModel(single_signal=MySingleSignalModel(value=1))
        )

        with self.assertRaises(AttributeError):
            self.assertFalse(form.Meta.custom_add_form)



from django.apps import AppConfig, apps
from django.contrib.auth import get_permission_codename
from django.core.checks import Warning

from ambient_toolbox.static_role_permissions.permissions import load_static_role_permissions


def collect_model_permissions(app_configs: list[AppConfig]) -> set[str]:
    """
    Go through all apps/models and collect default + custom permissions.
    """
    all_model_permissions = set()

    for app_config in app_configs:
        # iterate over all models in all apps
        for klass in app_config.get_models():
            opts = klass._meta
            app_label = app_config.label

            # collect default permissions
            for action in opts.default_permissions:
                codename = get_permission_codename(action, opts)
                all_model_permissions.add(f"{app_label}.{codename}")
            # collect custom permissions
            for codename, _ in opts.permissions:
                all_model_permissions.add(f"{app_label}.{codename}")

    return all_model_permissions


def check_permissions_against_models(app_configs: list[AppConfig] | None = None, **kwargs) -> list:
    """
    Check whether static role permissions exist in models.
    We do this to ensure full compatibility with Django's permission system.
    """
    if not app_configs:
        app_configs = apps.get_app_configs()

    errors = []
    model_permissions_set = collect_model_permissions(app_configs)
    role_permissions_dict = load_static_role_permissions()

    # check static permissions against model permissions
    for role_name, role_permissions_set in role_permissions_dict.items():
        for permission in role_permissions_set:
            if permission not in model_permissions_set:
                errors.append(
                    Warning(
                        "Permission does not exist in any model.",
                        obj=f"'{permission}' (Role '{role_name}')",
                    )
                )
                break

    return errors



from rest_framework.serializers import ModelSerializer


class BaseModelSerializer(ModelSerializer):
    def validate(self, data):
        """
        Call Model's clean() method to ensure model-level-validation
        :param data:
        :return:
        """
        cleaned_data = super().validate(data)
        instance = self.Meta.model(**cleaned_data)
        instance.clean()
        return cleaned_data


class CommonInfoSerializer(BaseModelSerializer):
    """
    A serializer that should be used for all models that extend "CommonInfo". It adds the data for
    `lastmodified_by` and `created_by` when saving a model through the serializer.
    This cannot be done in the model's `save` function, since the request is required.
    """

    def validate(self, data):
        data = super().validate(data)
        request = self.context.get("request", None)

        if request.user:
            data["lastmodified_by"] = request.user
            if not self.instance:
                # If this is a new instance, set created_by
                data["created_by"] = request.user

        return data



from django.contrib import admin
from django.core.exceptions import PermissionDenied


class AdminViewMixin:
    """
    Mixin to provide a custom view with all the attributes it needs to look like a regular django admin page.
    """

    model = None
    admin_page_title = ""

    def has_view_permission(self, user, **kwargs) -> bool:
        """
        Custom admin views are prone to be left open for all users.
        This method is called in the `dispatch()` and ensures the user has the required permissions to access
        this view. Defaults to the `is_superuser` flag of the django user.
        Can be overwritten if a different permission logic is required.
        """
        return user.is_superuser

    def dispatch(self, request, *args, **kwargs):
        """
        Runs a custom validation to ensure user has the permissions to access this page before running the default
        dispatch logic.
        """
        if self.has_view_permission(request.user):
            return super().dispatch(request, *args, **kwargs)
        raise PermissionDenied

    def get_admin_site(self):
        """
        Returns the set admin site. Can be overwritten if needed.
        """
        return admin.site

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        admin_site = self.get_admin_site()
        context.update(
            {
                "site_header": admin_site.site_header,
                "site_title": admin_site.site_title,
                "title": self.admin_page_title,
                "name": self.admin_page_title,
                "original": self.admin_page_title,
                "is_nav_sidebar_enabled": True,
                "available_apps": admin.site.get_app_list(self.request),
                "opts": {
                    "app_label": self.model._meta.app_label,
                    "verbose_name": self.model._meta.verbose_name,
                    "verbose_name_plural": self.model._meta.verbose_name_plural,
                    "model_name": self.model._meta.model_name,
                    "app_config": {
                        "verbose_name": self.model._meta.app_config.verbose_name,
                    },
                },
                "has_permission": admin_site.has_permission(request=self.request),
            }
        )
        return context



from unittest import mock

from django.core.mail import EmailMultiAlternatives
from django.core.mail.backends.smtp import EmailBackend
from django.test import TestCase, override_settings

from ambient_toolbox.mail.backends.whitelist_smtp import WhitelistEmailBackend


@override_settings(
    EMAIL_BACKEND="ambient_toolbox.mail.backends.whitelist_smtp.WhitelistEmailBackend",
    EMAIL_BACKEND_DOMAIN_WHITELIST=["valid.domain"],
    EMAIL_BACKEND_REDIRECT_ADDRESS="%s@testuser.valid.domain",
)
class MailBackendWhitelistBackendTest(TestCase):
    def test_whitify_mail_addresses_replace(self):
        email_1 = "albertus.magnus@example.com"
        email_2 = "thomas_von_aquin@example.com"
        processed_list = WhitelistEmailBackend.whitify_mail_addresses(mail_address_list=[email_1, email_2])

        self.assertEqual(len(processed_list), 2)
        self.assertEqual(processed_list[0], "albertus.magnus_example.com@testuser.valid.domain")
        self.assertEqual(processed_list[1], "thomas_von_aquin_example.com@testuser.valid.domain")

    def test_whitify_mail_addresses_whitelisted_domain(self):
        email = "platon@valid.domain"
        processed_list = WhitelistEmailBackend.whitify_mail_addresses(mail_address_list=[email])

        self.assertEqual(len(processed_list), 1)
        self.assertEqual(processed_list[0], email)

    @override_settings(EMAIL_BACKEND_REDIRECT_ADDRESS="")
    def test_whitify_mail_addresses_no_redirect_configured(self):
        email = "sokrates@example.com"
        processed_list = WhitelistEmailBackend.whitify_mail_addresses(mail_address_list=[email])

        self.assertEqual(len(processed_list), 0)

    def test_process_recipients_regular(self):
        mail = EmailMultiAlternatives(
            "Test subject", "Here is the message.", "from@example.com", ["to@example.com"], connection=None
        )

        backend = WhitelistEmailBackend()
        message_list = backend._process_recipients([mail])
        self.assertEqual(len(message_list), 1)
        self.assertEqual(message_list[0].to, ["to_example.com@testuser.valid.domain"])

    @mock.patch.object(EmailBackend, "send_messages")
    @mock.patch.object(WhitelistEmailBackend, "_process_recipients")
    def test_send_messages_process_recipients_called(self, mocked_process_recipients, *args):
        backend = WhitelistEmailBackend()
        backend.send_messages([])

        mocked_process_recipients.assert_called_once_with([])

    @mock.patch.object(EmailBackend, "send_messages")
    def test_send_messages_super_called(self, mocked_send_messages):
        backend = WhitelistEmailBackend()
        backend.send_messages([])

        mocked_send_messages.assert_called_once_with([])



from django.contrib.auth.models import AnonymousUser, User
from django.http import HttpResponse
from django.test import RequestFactory, TestCase
from django.views import generic

from ambient_toolbox.tests.mixins import ClassBasedViewTestMixin


class TestView(generic.TemplateView):
    def get(self, request, *args, **kwargs):
        return HttpResponse(status=202)

    def post(self, request, *args, **kwargs):
        return HttpResponse(status=203)

    def delete(self, request, *args, **kwargs):
        return HttpResponse(status=204)


class ClassBasedViewTestMixinTest(ClassBasedViewTestMixin, TestCase):
    view_class = TestView

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        factory = RequestFactory()
        cls.request = factory.get("/admin")
        cls.user = User.objects.create(username="my-username")

    def test_authentication_user_given(self):
        self._authentication(self.request, self.user), self.user
        self.assertEqual(self.request.user, self.user)

    def test_authentication_no_user_given(self):
        self._authentication(self.request, None), self.user
        self.assertEqual(self.request.user, AnonymousUser())

    def test_get_response_regular(self):
        response = self._get_response(method="get", user=self.user, data={"my_data": 42})
        self.assertIsInstance(response, HttpResponse)
        self.assertEqual(response.status_code, 202)

    def test_get_regular(self):
        response = self.get()
        self.assertIsInstance(response, HttpResponse)
        self.assertEqual(response.status_code, 202)

    def test_post_regular(self):
        response = self.post()
        self.assertIsInstance(response, HttpResponse)
        self.assertEqual(response.status_code, 203)

    def test_delete_regular(self):
        response = self.delete()
        self.assertIsInstance(response, HttpResponse)
        self.assertEqual(response.status_code, 204)



from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db import transaction

from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration


class PermissionSetupService:
    group_declaration: GroupPermissionDeclaration
    dry_run: bool

    def __init__(self, group_declaration: GroupPermissionDeclaration, dry_run: bool = False) -> None:
        super().__init__()

        self.group_declaration = group_declaration
        self.dry_run = dry_run

    @transaction.atomic
    def process(self) -> (list[Permission], list[Permission]):
        # Fetch or create group
        group, created = Group.objects.get_or_create(name=self.group_declaration.name)

        # Declare lists where we'll sort in the permissions
        defined_permission_list = []
        new_permissions = []
        removed_permissions = []

        # Iterate all in fixtures defined permission models...
        for permission_declaration in self.group_declaration.permission_list:
            # Iterate all fixture permissions...
            for codename in permission_declaration.codename_list:
                # Instantiate permission as ORM object
                try:
                    # Cast strings to lower cases to avoid issues with SQLite
                    new_permission = Permission.objects.get_by_natural_key(
                        app_label=permission_declaration.app_label.lower(),
                        codename=codename.lower(),
                        model=permission_declaration.model.lower(),
                    )
                except Permission.DoesNotExist as e:
                    raise ValueError(f'Invalid permission "{permission_declaration.model}.{codename}" declared.') from e
                except ContentType.DoesNotExist as e:
                    raise ValueError(
                        f'Invalid content type "{permission_declaration.app_label}'
                        f'.{permission_declaration.model}" declared.'
                    ) from e

                # Add permission object to list
                if new_permission in defined_permission_list:
                    raise ValueError(f"Permission {new_permission} declared twice.")
                defined_permission_list.append(new_permission)

                # Check if permission is already set in the group
                if new_permission not in group.permissions.all():
                    new_permissions.append(new_permission)

        # Check which permissions were removed for the given group
        for existing_permission in group.permissions.all():
            if existing_permission not in defined_permission_list:
                removed_permissions.append(existing_permission)  # noqa: PERF401

        if not self.dry_run:
            # Persist changes on removed permissions
            if len(removed_permissions):
                group.permissions.remove(*removed_permissions)

            # Persist changes on added permissions
            if len(new_permissions):
                group.permissions.add(*new_permissions)

        # Return changes
        return new_permissions, removed_permissions



import json
import os
import re
import subprocess
import sys
from difflib import ndiff
from http import HTTPStatus
from typing import Optional

import httpx


class CoverageService:
    """
    Class to be used in the gitlab-ci to ensure pipeline test coverage is not dropping on commit
    """

    def __init__(self) -> None:
        super().__init__()

        # Get ENV variables
        self.current_pipeline_id: int = int(os.environ.get("CI_PIPELINE_ID"))
        self.base_api_url: str = os.environ.get("CI_API_V4_URL")
        self.current_branch: str = os.environ.get("CI_COMMIT_REF_NAME")
        self.token: str = os.environ.get("GITLAB_CI_COVERAGE_PIPELINE_TOKEN")
        self.project_id: int = int(os.environ.get("CI_PROJECT_ID"))
        self.job_name: str = os.environ.get("CI_COVERAGE_JOB_NAME", "")
        self.target_branch: str = os.environ.get("GITLAB_CI_COVERAGE_TARGET_BRANCH", "develop")
        self.pipelines_url = (
            f"{self.base_api_url}/projects/{self.project_id}/pipelines?ref={self.target_branch}&status=success"
        )
        self.pipelines_url_with_token = f"{self.pipelines_url}&private_token={self.token}"

        self.disable_coverage: bool = self.get_disable_coverage(os.environ.get("GITLAB_CI_DISABLE_COVERAGE", "0"))

    @staticmethod
    def get_disable_coverage(disable_env: str) -> bool:
        disable_coverage = disable_env

        if disable_coverage.lower() == "true":
            return True
        elif disable_coverage.lower() == "false":
            return False
        elif disable_coverage.isdigit():
            return bool(int(disable_coverage))
        else:
            return bool(disable_coverage)

    def get_latest_target_branch_commit_sha(self) -> str:
        """
        Get the latest commit which is in the current branch and the target compare branch.
        """
        result = subprocess.run(  # noqa: PLW1510
            ["git", "merge-base", "--fork-point", f"origin/{self.target_branch}"],
            stdout=subprocess.PIPE,
        )
        return result.stdout.decode("utf-8").strip()

    def get_pipeline_id_by_commit_sha(self, sha: str) -> Optional[int]:
        pipeline_url = f"{self.pipelines_url_with_token}&sha={sha}"
        response = httpx.get(pipeline_url)
        status_code = response.status_code

        if status_code == HTTPStatus.OK:
            pipelines = json.loads(response.content)
            if pipelines and len(pipelines) > 0:
                return pipelines[0].get("id", None)
            else:
                print("\n### ERROR: No pipelines found for SHA1 ###\n")
                print(f"Pipeline URL: {self.pipelines_url}&sha={sha}")
                print(response.content)
        return None

    def get_coverage_from_pipeline(self, pipeline_id: int, job_name: str) -> (float, float):
        """
        Get coverage from given pipeline (by id)
        """
        jobs_url = f"{self.base_api_url}/projects/{self.project_id}/pipelines/{pipeline_id}/jobs"
        jobs_with_token_url = f"{jobs_url}?private_token={self.token}"

        print(f"Jobs-API-URL: {jobs_url}")
        jobs_response = httpx.get(jobs_with_token_url)
        jobs_status_code = jobs_response.status_code

        if jobs_status_code != HTTPStatus.OK:
            raise ConnectionError(f"Call to jobs api endpoint failed with status code {jobs_status_code}")

        jobs = json.loads(jobs_response.content)
        coverages = {
            job["name"]: {"id": job["id"], "coverage": float(job["coverage"]), "web_url": job["web_url"]}
            for job in jobs
            if job.get("coverage")
        }

        pipeline_url = f"{self.base_api_url}/projects/{self.project_id}/pipelines/{pipeline_id}"
        pipeline_with_token_url = f"{pipeline_url}?private_token={self.token}"
        pipeline_response = httpx.get(pipeline_with_token_url)
        pipeline_status_code = pipeline_response.status_code

        if pipeline_status_code != HTTPStatus.OK:
            raise ConnectionError(f"Call to pipeline api endpoint failed with status code {pipeline_status_code}")

        pipeline = json.loads(pipeline_response.content)
        coverages_total = float(pipeline["coverage"] if pipeline["coverage"] else 0.0)
        print(f"Pipeline-API-URL: {pipeline_url}")
        print(f'Pipeline-URL: {pipeline["web_url"]}')

        if job_name == "":
            print(
                "\033[91mATTN: No CI_COVERAGE_JOB_NAME provided, using Total Coverage and skipping Coverage Diff\033[0m"
            )
            return coverages_total, coverages_total, None

        coverage_job = coverages.get(job_name)

        print(f'Job-URL: {coverage_job["web_url"]}')

        job_url = f'{self.base_api_url}/projects/{self.project_id}/jobs/{coverage_job["id"]}/trace'
        job_with_token_url = f"{job_url}?private_token={self.token}"
        job_response = httpx.get(job_with_token_url)
        job_status_code = job_response.status_code

        if job_status_code != HTTPStatus.OK:
            raise ConnectionError(f"Call to job api endpoint failed with status code {job_status_code}")

        print(f"Job-Log-URL: {job_url}")

        job_log = re.search(
            r"Name\s+Stmts\s+Miss\s+Branch\s+BrPart\s+Cover\s+Missing.*files skipped due to complete coverage\.",
            job_response.content.decode("utf-8"),
            re.DOTALL | re.MULTILINE,
        )
        job_log_group = job_log.group() if job_log else None

        return coverage_job["coverage"] if coverage_job else 0.0, coverages_total, job_log_group

    @staticmethod
    def color_text(sign: int, prefix: str, target: float, current: float, diff: float):
        """
        function to return colored in text according to the template:
        "{Total|Job} coverage {change_text} from {target}% to {current}% (Diff: {diff}%)"
        Red color: coverage dropped
        White/No color: coverage unchanged
        Green color: coverage climbed
        :param sign: numeric value of the coverage diff sign
        :param prefix: text prefix (i.e.: Total coverage, Job Coverage)
        :param target: target percentage
        :param current: current percentage
        :param diff: difference between current and target percentage
        :return: fully assembled and colored summary text
        """
        change = {
            -1: {"text": "dropped", "color": "\033[91m"},
            0: {"text": "unchanged", "color": ""},
            1: {"text": "climbed", "color": "\033[92m"},
        }
        return (
            f'{change[sign]["color"]} {prefix} {change[sign]["text"]} '
            f'from {target:2.2f}% to {current:2.2f}% (Diff: {diff:2.2f}%).\033[0m'
        )

    @staticmethod
    def print_diff(target_job_log, current_job_log):
        """
        Print a diff between the coverage reports of Current and Target branch
        """
        diff = ndiff(target_job_log.splitlines(keepends=True), current_job_log.splitlines(keepends=True))
        print("\n############################## Coverage Diff ##############################")
        print("# \033[91m- Target Branch\033[0m                                                         #")
        print("# \033[92m+ Current Branch\033[0m                                                        #")
        print("###########################################################################")
        for _idx, line in enumerate(diff):
            match = re.match(
                r"^\s*-+\s*$|"
                r"^\s*Name\s+Stmts\s+Miss\s+Branch\s+BrPart\s+Cover\s+Missing|"  # first line of the report
                r"^.*files skipped due to complete coverage.$|"  # Final line of the report
                r"^[+\-?]",  # Line starts with +,-,$ to indicate changes
                line,
            )
            if match:
                if line[0] == "-":
                    print("\033[91m", end="")
                if line[0] == "+":
                    print("\033[92m", end="")
                print(line, end="")
                if line[0] in ["+", "-"]:
                    print("\033[0m", end="")

        print("\n###########################################################################")
        print("# \033[91m- Target Branch\033[0m                                                         #")
        print("# \033[92m+ Current Branch\033[0m                                                        #")
        print("############################## Coverage Diff ##############################\n")

    def process(self):
        """
        Compare coverage from target branch (latest develop) with the current one.
        At first, we try to get a successfully finished pipeline for the "self.target_branch" (usually "develop")
        """
        # Check, if coverage is supposed to run. If not, inform the user and return early.
        if self.disable_coverage:
            print("Coverage was skipped!")
            sys.exit(0)

        print("\n###########################################################################\n")
        print("DEBUG INFO:")

        # Get the latest commit SHA which is also in develop
        commit_sha = self.get_latest_target_branch_commit_sha()

        # Try to find the latest successful pipeline for "TARGET_BRANCH" where our SHA was in
        print("Trying base branch for comparison.")
        target_pipeline_id = None
        if commit_sha:
            print(f'Found latest target branch commit SHA "{commit_sha}".')
            target_pipeline_id = self.get_pipeline_id_by_commit_sha(commit_sha)
            print(f"Target branch pipeline ID identified: {target_pipeline_id}.")

        # Get target pipeline id (from develop branch) if we were not successful the first time
        if not target_pipeline_id:
            print("Didn't work. Using default branch for comparison.")
            response = httpx.get(self.pipelines_url_with_token)
            status_code = response.status_code
            print(f"Pipelines-API-URL: {self.pipelines_url}")

            # Ensure call did not go sideways
            if status_code != HTTPStatus.OK:
                raise ConnectionError(f"Call to global pipeline api endpoint failed with status code {status_code}")

            # Read target pipeline ID from content
            try:
                target_pipeline_id = json.loads(response.content)[0]["id"]
            except IndexError:
                # This happens when there are zero `target_branch` pipelines
                target_pipeline_id = 0
            print(f"Default branch pipeline ID identified: {target_pipeline_id}.")

        # Get coverage from target pipeline
        print(f"Target Pipeline ID: {target_pipeline_id}")
        target_job_coverage, target_total_coverage, target_job_log = self.get_coverage_from_pipeline(
            target_pipeline_id, self.job_name
        )

        # Get coverage from this pipeline
        print(f"Current pipeline ID: {self.current_pipeline_id}")
        current_job_coverage, current_total_coverage, current_job_log = self.get_coverage_from_pipeline(
            self.current_pipeline_id, self.job_name
        )

        if target_job_log is None or current_job_log is None:
            print("\n\n\033[91m***************************************************************************\033[0m")
            print("        \033[91m**/!\\** Coverage log not found. Skipping diff. **/!\\**\033[0m             ")
            print("\033[91m***************************************************************************\033[0m\n\n")
        else:
            self.print_diff(target_job_log, current_job_log)

        # numeric value of the coverage diff sign
        sign_job_coverage = (current_job_coverage > target_job_coverage) - (current_job_coverage < target_job_coverage)
        sign_total_coverage = (current_total_coverage > target_total_coverage) - (
            current_total_coverage < target_total_coverage
        )

        # difference between current and target coverage
        diff_job_coverage = current_job_coverage - target_job_coverage
        diff_total_coverage = current_total_coverage - target_total_coverage

        coverage = {
            "total": {
                "target": target_total_coverage,
                "current": current_total_coverage,
                "sign": sign_total_coverage,
                "diff": diff_total_coverage,
                "prefix": "Total coverage",
            },
            "job": {
                "target": target_job_coverage,
                "current": current_job_coverage,
                "sign": sign_job_coverage,
                "diff": diff_job_coverage,
                "prefix": "Job coverage",
            },
        }

        # Print results
        print(
            self.color_text(
                coverage["total"]["sign"],
                coverage["total"]["prefix"],
                coverage["total"]["target"],
                coverage["total"]["current"],
                coverage["total"]["diff"],
            )
        )
        print(
            self.color_text(
                coverage["job"]["sign"],
                coverage["job"]["prefix"],
                coverage["job"]["target"],
                coverage["job"]["current"],
                coverage["job"]["diff"],
            )
        )

        if coverage["job"]["sign"] == -1:
            sys.exit(1)



from django.contrib.auth.models import Group, Permission
from django.test import TestCase

from ambient_toolbox.permissions.fixtures.declarations import GroupPermissionDeclaration, PermissionModelDeclaration
from ambient_toolbox.permissions.fixtures.services import PermissionSetupService


class PermissionSetupServiceTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.group, created = Group.objects.get_or_create(name="my_group")

        cls.permission_view = Permission.objects.get_by_natural_key(
            app_label="testapp", codename="view_mysinglesignalmodel", model="mysinglesignalmodel"
        )
        cls.permission_change = Permission.objects.get_by_natural_key(
            app_label="testapp", codename="change_mysinglesignalmodel", model="mysinglesignalmodel"
        )

    def test_init_regular(self):
        group_declaration = GroupPermissionDeclaration(name="my_group", permission_list=[])
        service = PermissionSetupService(group_declaration=group_declaration)

        self.assertEqual(service.group_declaration, group_declaration)
        self.assertFalse(service.dry_run)

    def test_process_add_permission(self):
        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["change_mysinglesignalmodel"], model="mysinglesignalmodel"
                )
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration)
        new_permissions, removed_permissions = service.process()

        self.assertEqual(new_permissions, [self.permission_change])
        self.assertEqual(removed_permissions, [])

        self.group.refresh_from_db()
        self.assertEqual(self.group.permissions.count(), 1)
        self.assertEqual(self.group.permissions.all().first(), self.permission_change)

    def test_process_remove_permission(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(name="my_group", permission_list=[])

        service = PermissionSetupService(group_declaration=group_declaration)
        new_permissions, removed_permissions = service.process()

        self.assertEqual(new_permissions, [])
        self.assertEqual(removed_permissions, [self.permission_view])

        self.group.refresh_from_db()
        self.assertFalse(self.group.permissions.exists())

    def test_process_no_changes_permission(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["view_mysinglesignalmodel"], model="mysinglesignalmodel"
                )
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration)
        new_permissions, removed_permissions = service.process()

        self.assertEqual(new_permissions, [])
        self.assertEqual(removed_permissions, [])

        self.group.refresh_from_db()
        self.assertEqual(self.group.permissions.count(), 1)
        self.assertEqual(self.group.permissions.all().first(), self.permission_view)

    def test_process_invalid_permission(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["invalid_permission"], model="mysinglesignalmodel"
                )
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration)
        with self.assertRaisesMessage(
            ValueError, 'Invalid permission "mysinglesignalmodel.invalid_permission" declared.'
        ):
            service.process()

    def test_process_invalid_app(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="invalid_app", codename_list=["view_mysinglesignalmodel"], model="mysinglesignalmodel"
                )
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration)
        with self.assertRaisesMessage(ValueError, 'Invalid content type "invalid_app.mysinglesignalmodel" declared.'):
            service.process()

    def test_process_invalid_model(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["view_mysinglesignalmodel"], model="invalid_model"
                )
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration)
        with self.assertRaisesMessage(ValueError, 'Invalid content type "testapp.invalid_model" declared.'):
            service.process()

    def test_process_duplicated_permission_declaration(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["view_mysinglesignalmodel"], model="mysinglesignalmodel"
                ),
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["view_mysinglesignalmodel"], model="mysinglesignalmodel"
                ),
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration)
        with self.assertRaisesMessage(ValueError, f"Permission {self.permission_view} declared twice."):
            service.process()

    def test_process_dry_run_not_persisting(self):
        self.group.permissions.add(self.permission_view)

        group_declaration = GroupPermissionDeclaration(
            name="my_group",
            permission_list=[
                PermissionModelDeclaration(
                    app_label="testapp", codename_list=["change_mysinglesignalmodel"], model="mysinglesignalmodel"
                )
            ],
        )

        service = PermissionSetupService(group_declaration=group_declaration, dry_run=True)
        new_permissions, removed_permissions = service.process()

        self.assertEqual(new_permissions, [self.permission_change])
        self.assertEqual(removed_permissions, [self.permission_view])

        self.group.refresh_from_db()
        self.assertEqual(self.group.permissions.count(), 1)
        self.assertEqual(self.group.permissions.all().first(), self.permission_view)



import warnings
from typing import TYPE_CHECKING, Callable

from .current_request import CurrentRequestMiddleware

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse


class CurrentUserMiddleware(CurrentRequestMiddleware):
    """
    Middleware which stores the current user in a thread-safe manner.

    This middleware is deprecated, use CurrentRequestMiddleware instead.
    """

    # Tech Note: This class is now a mere alias for CurrentRequestMiddleware to
    # provide backward compatibility. This whole module can be removed as part
    # of one of the next major releases, then fully dropping support for
    # CurrentUserMiddleware.

    def __init__(self, get_response: Callable[["HttpRequest"], "HttpResponse"]):
        warnings.warn(
            "CurrentUserMiddleware is deprecated. Use CurrentRequestMiddleware instead.",
            category=DeprecationWarning,
            stacklevel=1,
        )
        super().__init__(get_response)



from django.test import TestCase

from testapp.models import MyPermissionModelMixin


class PermissionModelMixinTest(TestCase):
    def test_meta_managed_false(self):
        self.assertFalse(MyPermissionModelMixin.Meta.managed)



import os
import subprocess

from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """
    Validates for all active languages (settings.LANGUAGES) the integrity of the PO translation files.
    The following cases are covered:
    * Fuzzy translations aren't allowed
    * Commented-out translations aren't allowed
    * Validate, that `manage.py makemessages` has been called before committing
    * Validate, that all translations were actually translated
    """

    def handle(self, *args, **options):
        for lang, label in settings.LANGUAGES:
            print(f'Check language "{label}"...')

            if not os.path.isfile(f"./locale/{lang}/LC_MESSAGES/django.po"):
                print("> Skipping language due to missing PO file.")
                continue

            # Check for fuzzy translations
            print("> Check for fuzzy translations")
            output = subprocess.call(f'grep -q "#, fuzzy" ./locale/{lang}/LC_MESSAGES/django.po', shell=True)
            if output == 0:
                raise Exception("Please remove all fuzzy translations in your translation files.")
            else:
                print("OK.")

            # Check for left-over translations
            print("> Check for left-over translations")
            output = subprocess.call(f'grep -q "#~" ./locale/{lang}/LC_MESSAGES/django.po', shell=True)
            if output == 0:
                raise Exception("Please remove all commented-out translations in your translation files.")
            else:
                print("OK.")

            # Check if "makemessages" detects new translations
            print('> Check if "makemessages" detects new translations')
            subprocess.call(f"python manage.py create_translation_file --lang {lang}", shell=True)
            print("OK.")

            # Checking for differences in translation file
            print("> Checking for differences in translation file")
            output = subprocess.call(
                r"git diff --ignore-matching-lines=POT-Creation-Date --ignore-matching-lines=# --exit-code locale/",
                shell=True,
            )
            if output > 0:
                raise Exception(
                    "It seems you have forgotten to update your translation file before pushing your changes."
                )
            else:
                print("OK.")

            # Check if all translation strings have been translated
            print("> Check if all translation strings have been translated")
            output = subprocess.call(
                f"msgattrib --untranslated ./locale/{lang}/LC_MESSAGES/django.po | exit `wc -c`", shell=True
            )
            if output > 0:
                raise Exception("You have untranslated strings in your translation files.")
            else:
                print("OK.")



from typing import Optional

from django.contrib.auth.models import AbstractUser
from django.http import HttpResponseRedirect
from django.urls import resolve, reverse


class AdminCreateFormMixin:
    """
    Mixin to easily use a different form for the create case (in comparison to "edit") in the django admin
    Logic copied from `django.contrib.auth.admin.UserAdmin`
    """

    add_form = None

    def get_form(self, request, obj=None, **kwargs):
        defaults = {}
        if obj is None:
            defaults["form"] = self.add_form
        defaults.update(kwargs)
        return super().get_form(request, obj, **defaults)


class AdminNoInlinesForCreateMixin:
    """
    Avoid showing any inline admins when being in the create case of a record
    """

    def get_inline_instances(self, request, obj=None):
        if obj is None:
            return []
        return super().get_inline_instances(request, obj)


class AdminRequestInFormMixin:
    """
    Mixin to add the current request to a form used in the django admin
    """

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.request = request
        return form


class FetchParentObjectInlineMixin:
    """
    Fetches the parent object via the URL resolver and makes it available throughout the entire class.
    Attention: Use only in inline admin classes.
    """

    parent_object = None

    @staticmethod
    def _resolve_url(request):
        return resolve(request.path_info)

    def get_parent_object_from_request(self, request):
        resolved = self._resolve_url(request)
        if resolved.kwargs:
            return self.parent_model.objects.get(pk=resolved.kwargs.get("object_id", None))
        return None

    def get_formset(self, request, obj=None, **kwargs):
        # Fetch parent object - cannot be done in the init method because request is not available there
        self.parent_object = self.get_parent_object_from_request(request)
        return super().get_formset(request, obj, **kwargs)


class FetchObjectMixin:
    """
    Fetches the current object via the URL resolver and makes it available throughout the entire class.
    """

    def get_object_from_request(self, request):
        resolved = resolve(request.path_info)
        if resolved.kwargs:
            return self.model.objects.get(pk=resolved.kwargs.get("object_id", None))
        return None


class CommonInfoAdminMixin:
    """
    Mixin to be used in a django model admin class.
    Sets all four `CommonInfo` attributes to "readonly" and sets the creator / last modifier on form save.
    """

    def get_readonly_fields(self, request, obj=None):
        """
        Set the fields CommonInfo handles to readonly to avoid users fiddling around with them.
        """
        return (
            *super().get_readonly_fields(request, obj),
            "created_by",
            "lastmodified_by",
            "created_at",
            "lastmodified_at",
        )

    def get_user_obj(self, request) -> Optional[AbstractUser]:
        """
        Fetches the user object from the current request.
        Can be overwritten when a different user object is required.
        """
        return request.user

    def save_form(self, request, form, change):
        if form.instance and request.user:
            if not form.instance.pk:
                form.instance.created_by = self.get_user_obj(request=request)
            form.instance.lastmodified_by = self.get_user_obj(request=request)

        return super().save_form(request, form, change)


class DeactivatableChangeViewAdminMixin:
    """
    Mixin to be used in model admins to disable the detail page / change view.
    """

    enable_change_view = True

    def can_see_change_view(self, request) -> bool:
        """
        A method that determines if the change view is disabled or visible.
        """
        return self.enable_change_view

    def get_list_display_links(self, request, list_display):
        """
        When we don't want to show the change view, there is no need for having a link to it
        """
        if not self.can_see_change_view(request=request):
            return None
        return super().get_list_display_links(request, list_display)

    def change_view(self, request, *args, **kwargs):
        """
        The 'change' admin view for this model.

        We override this to redirect back to the changelist unless the view is
        specifically enabled by the "enable_change_view" property.
        """
        if self.can_see_change_view(request=request):
            return super().change_view(request, *args, **kwargs)
        else:
            opts = self.model._meta
            url = reverse(
                "admin:{app}_{model}_changelist".format(
                    app=opts.app_label,
                    model=opts.model_name,
                )
            )
            return HttpResponseRedirect(url)



from django.contrib import admin


class ReadOnlyTabularInline(admin.TabularInline):
    """
    Class for being extended by TabularInline-classes.
    Disables all create, delete or edit functionality in the tabular inline admin.
    """

    can_delete = False

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def get_readonly_fields(self, request, obj=None):
        result = list(
            set(
                [field.name for field in self.opts.local_fields]
                + [field.name for field in self.opts.local_many_to_many]
            )
        )
        result.remove("id")
        return result



from rest_framework import serializers


class RecursiveField(serializers.Serializer):
    """
    Custom field for using the same serializer as a field. Useful for foreign keys to the same model.
    """

    def update(self, instance, validated_data):
        super().update(instance, validated_data)

    def create(self, validated_data):
        super().create(validated_data)

    def to_representation(self, value):
        # If the field is used with `many=True` we need to go one more parent level up to get the "real"
        # parent serializer.
        # Explanation: With `many=True` DRF creates an intermediate `ListSerializer`. It has `many=True`, so we'll end
        # up in the first if-case. If we do not use `many=True`, the "many" attribute is not set.
        if getattr(self.parent, "many", False):
            parent = self.parent.parent
        else:
            parent = self.parent
        serializer = parent.__class__(value, context=self.context)
        return serializer.data



import json
from typing import Union


class HtmxResponseMixin:
    """
    View mixin to be able to simply set "HX-Redirect" and "HX-Trigger" for using HTMX in the frontend.
    "hx_redirect_url": Takes a reverse_lazy URL to a valid Django URL
    "hx_trigger": Takes either a string "updateComponentX" or a dictionary with a key-value-pair, where the key is
    the signal and the value is a parameter passed to the frontend. If you don't need the value, set it to None.
    """

    hx_redirect_url: str = None
    hx_trigger: Union[str, dict[str, str]] = None

    def dispatch(self, request, *args, **kwargs):
        response = super().dispatch(request, *args, **kwargs)

        # Get attributes
        hx_redirect_url = self.get_hx_redirect_url()
        hx_trigger = self.get_hx_trigger()

        # Set redirect header if set
        if hx_redirect_url:
            response["HX-Redirect"] = hx_redirect_url

        # Set trigger header if set
        if isinstance(hx_trigger, dict):
            response["HX-Trigger"] = json.dumps(hx_trigger)
        elif isinstance(hx_trigger, str):
            response["HX-Trigger"] = hx_trigger

        # Return augmented response
        return response

    def get_hx_redirect_url(self):
        """
        Getter for "hx_redirect_url" to be able to work with dynamic data
        """
        return self.hx_redirect_url

    def get_hx_trigger(self):
        """
        Getter for "hx_trigger" to be able to work with dynamic data
        """
        return self.hx_trigger



from rest_framework.routers import DefaultRouter

from testapp.api.views import MySingleSignalModelViewSet

model_router = DefaultRouter()
model_router.register(r"my-single-signal-model", MySingleSignalModelViewSet, basename="my-single-signal-model")



from django import template

register = template.Library()


@register.filter(name="multiply")
def multiply(value, arg):
    """
    Multiplies the arg and the value
    """
    if value:
        value = f"{value}"
        if isinstance(value, str) and len(value) > 0:
            return float(value.replace(",", ".")) * float(arg)

    return None


@register.filter(name="subtract")
def subtract(value, arg):
    """
    Subtracts the arg from the value
    """
    value = value if value is not None else 0
    arg = arg if arg is not None else 0
    return int(value) - int(arg)


@register.filter(name="divide")
def divide(value, arg):
    """
    Divides the value by the arg
    """
    if value:
        return value / arg
    else:
        return None


@register.filter(name="to_int")
def to_int(value):
    """
    Parses a string to int value
    """
    try:
        return int(value)
    except ValueError:
        return 0


@register.filter(name="currency")
def currency(value):
    """
    Converts the number to an €-amount
    """
    if value:
        return (("%.2f" % round(value, 2)) + "€").replace(".", ",")
    else:
        return "-"



from django.test import TestCase

from ambient_toolbox.selectors.permission import AbstractUserSpecificSelectorMixin
from testapp.models import ModelWithSelector


class AbstractUserSpecificSelectorMixinTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.mixin = AbstractUserSpecificSelectorMixin()

    def test_visible_for_method_available(self):
        with self.assertRaises(NotImplementedError):
            self.mixin.visible_for(user_id=-1)

    def test_editable_for_method_available(self):
        with self.assertRaises(NotImplementedError):
            self.mixin.editable_for(user_id=-1)

    def test_deletable_for_method_available(self):
        with self.assertRaises(NotImplementedError):
            self.mixin.deletable_for(user_id=-1)


class GloballyVisibleSelectorMixinTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.obj = ModelWithSelector.objects.create(value=1)

    def test_visible_for_method_available(self):
        qs = ModelWithSelector.selectors.visible_for(user_id=-1)

        self.assertEqual(qs.count(), 1)
        self.assertIn(self.obj, qs)

    def test_editable_for_method_available(self):
        qs = ModelWithSelector.selectors.editable_for(user_id=-1)

        self.assertEqual(qs.count(), 1)
        self.assertIn(self.obj, qs)

    def test_deletable_for_method_available(self):
        qs = ModelWithSelector.selectors.deletable_for(user_id=-1)

        self.assertEqual(qs.count(), 1)
        self.assertIn(self.obj, qs)



import datetime

import pytz
from django.test import TestCase

from ambient_toolbox.utils.string import (
    date_to_string,
    datetime_to_string,
    distinct,
    encode_to_xml,
    float_to_string,
    number_to_string,
    slugify_file_name,
    smart_truncate,
    string_or_none_to_string,
)


class UtilsStringTest(TestCase):
    def test_distinct_regular(self):
        not_distinct_list = ["Beer", "Wine", "Whiskey", "Beer"]
        distinct_list = distinct(not_distinct_list)

        self.assertEqual(len(distinct_list), 3)
        self.assertIn("Beer", distinct_list)
        self.assertIn("Whiskey", distinct_list)
        self.assertIn("Wine", distinct_list)

    def test_slugify_file_name_regular(self):
        filename = "hola and hello.txt"
        slug = slugify_file_name(filename)
        self.assertEqual(slug, "hola_and_hello.txt")

    def test_slugify_file_name_nothing_to_slugify(self):
        filename = "hola.txt"
        slug = slugify_file_name(filename)
        self.assertEqual(slug, filename)

    def test_slugify_file_name_max_length(self):
        filename = "a very long filename.txt"
        slug = slugify_file_name(filename, 6)
        self.assertEqual(slug, "a_very.txt")

    def test_smart_truncate_in_word(self):
        my_sentence = "I am a very interesting sentence."
        truncated_str = smart_truncate(my_sentence, 10)
        self.assertEqual(truncated_str, "I am a...")

    def test_smart_truncate_after_word(self):
        my_sentence = "I am a very interesting sentence."
        truncated_str = smart_truncate(my_sentence, 14)
        self.assertEqual(truncated_str, "I am a very...")

    def test_smart_truncate_changed_postfix(self):
        my_sentence = "I am a very interesting sentence."
        truncated_str = smart_truncate(my_sentence, 10, "[...]")
        self.assertEqual(truncated_str, "I am a[...]")

    def test_smart_truncate_not_cutting_on_too_short_strings(self):
        my_sentence = "I am a very interesting sentence."
        truncated_str = smart_truncate(my_sentence, 100, "---")
        self.assertEqual(truncated_str, my_sentence)

    def test_smart_truncate_no_text(self):
        truncated_str = smart_truncate(None, 100, "---")
        self.assertEqual(truncated_str, "")

    def test_float_to_string_regular(self):
        self.assertEqual(float_to_string(5.61), "5,61")

    def test_float_to_string_value_replacement_not_used(self):
        self.assertEqual(float_to_string(4.41, "-"), "4,41")

    def test_float_to_string_no_value_replacement_used(self):
        self.assertEqual(float_to_string(None, "Heureka"), "Heureka")

    def test_float_to_string_value_greater_thousand(self):
        self.assertEqual(float_to_string(1234.56), "1234,56")

    def test_date_to_string_regular(self):
        self.assertEqual(date_to_string(datetime.date(2020, 9, 19)), "19.09.2020")

    def test_date_to_string_other_format(self):
        self.assertEqual(date_to_string(datetime.date(2020, 9, 19), str_format="%Y-%m-%d"), "2020-09-19")

    def test_date_to_string_replacement_undefined(self):
        self.assertEqual(date_to_string(None), "-")

    def test_date_to_string_replacement_defined(self):
        self.assertEqual(date_to_string(None, "no date"), "no date")

    def test_datetime_to_string_regular(self):
        self.assertEqual(datetime_to_string(datetime.datetime(2020, 9, 19, 8, tzinfo=pytz.UTC)), "19.09.2020 08:00")

    def test_datetime_to_string_other_format(self):
        self.assertEqual(
            datetime_to_string(datetime.datetime(2020, 9, 19, 8, tzinfo=pytz.UTC), str_format="%Y-%m-%d"), "2020-09-19"
        )

    def test_datetime_to_string_replacement_undefined(self):
        self.assertEqual(datetime_to_string(None), "-")

    def test_datetime_to_string_replacement_defined(self):
        self.assertEqual(datetime_to_string(None, "no date"), "no date")

    def test_number_to_string_regular(self):
        self.assertEqual(number_to_string(5.61, decimal_digits=2), "5.61")

    def test_number_to_string_value_replacement_not_used(self):
        self.assertEqual(number_to_string(4.41, decimal_digits=2, replacement="-"), "4.41")

    def test_number_to_string_no_value_replacement_used(self):
        self.assertEqual(number_to_string(None, replacement="Heureka"), "Heureka")

    def test_number_to_string_value_greater_thousand(self):
        self.assertEqual(number_to_string(1234.56, decimal_digits=2), "1,234.56")

    def test_number_to_string_int_value_no_digits(self):
        self.assertEqual(number_to_string(117), "117")

    def test_number_to_string_int_value_with_digits(self):
        self.assertEqual(number_to_string(117, decimal_digits=2), "117.00")

    def test_string_or_none_to_string_regular(self):
        my_str = "I am a string."
        self.assertEqual(string_or_none_to_string(my_str), my_str)

    def test_string_or_none_to_string_replacement_undefined(self):
        self.assertEqual(string_or_none_to_string(None), "-")

    def test_string_or_none_to_string_replacement_defined(self):
        self.assertEqual(string_or_none_to_string(None, "no value"), "no value")

    def test_encode_to_xml_regular(self):
        xml_str = "<tag>Something with an ampersand (&)</tag>"
        self.assertEqual(encode_to_xml(xml_str), "&lt;tag&gt;Something with an ampersand (&amp;)&lt;/tag&gt;")



from django.test import TestCase

from ambient_toolbox.permissions.fixtures.helpers import generate_default_permissions


class PermissionFixtureHelperTest(TestCase):
    def test_generate_default_permissions_regular(self):
        permission_list = generate_default_permissions("mymodel")

        self.assertEqual(len(permission_list), 4)
        self.assertIn("add_mymodel", permission_list)
        self.assertIn("change_mymodel", permission_list)
        self.assertIn("delete_mymodel", permission_list)
        self.assertIn("view_mymodel", permission_list)
